#############
# $Id: bustard.pm 12510 2011-02-01 15:49:31Z ajb $
# Created By: ts6
# Last Maintained By: $Author: ajb $
# Created On: 2009-09-18
# Last Changed On: $Date: 2011-02-01 15:49:31 +0000 (Tue, 01 Feb 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-24.2/lib/npg_pipeline/analysis/bustard.pm $

#----------------------------------------------------------------------

# Package to run the Illumina RTA post-run analysis pipeline, from
# .cif files through the creation of qseq.txt files. It expects to be
# invoked from harold.pm. It runs an initial (quick) step to create
# directories and makefiles (the old 'goat' step), then bsub's a bunch
# of 'make' jobs to do the actual processing (the old 'lsgmake' step).

#----------------------------------------------------------------------

package npg_pipeline::analysis::bustard;

use Moose;
####use MooseX::InsideOut;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12510 $ =~ /(\d+)/mxs; $r; };
use Cwd;      # for getcwd
use Perl6::Slurp;

extends q{npg_pipeline::base};
with q{npg_common::roles::run::long_info};
with q{npg_common::roles::run::fs_resource};
with q{npg_pipeline::roles::business::bustard_lsf_reqs};
with q{npg_common::roles::run::lane::tag_info};

Readonly::Scalar our $ACCESS_ALL => oct 777;                # octal value for chmod

=head1 NAME

npg_pipeline::analysis::bustard

=head1 VERSION

$LastChangedRevision: 12510 $

=head1 SYNOPSIS

  my $oAnalysisBustard = npg_pipeline::analysis::bustard->new({
    timestamp => $sTimeStamp,
  });

=head1 DESCRIPTION

Object module which knows how to create and submit analysis jobs using the pipeline

=head1 SUBROUTINES/METHODS

=head2 timestamp

a string containing the timestamp for this analysis. Required on construction

  my $sTimeStamp = $oAnalysisBustard->timestamp();

=cut

has q{timestamp} => (isa => q{Str}, is => q{ro}, required => 1);


=head2 illumina_pipeline

string containing path to the pipeline to be used, can be set on construction if you wish to pass in a different pipeline location

  my $sIlluminaPipeline = $oAnalysisBustard->illumina_pipeline();

=head2 submit_bustard

Run the Bustard setup step, which creates directories and makefiles
for post-run analysis. This is quick, so it's not bsub'd.
Returns an empty array

  my @EmptyArray = $oAnalysisBustard->submit_bustard();

=cut

# ----------------------------------------------------------------------

sub submit_bustard {

  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();

  $self->log("Running Bustard+Gerald makefile creation for run $id_run");

  $self->_check_runfolder($arg_refs);

  $arg_refs->{control_lane}  = $self->_pick_control_lane();
#### $arg_refs->{control_lane}  = q{};
  $self->log(q{Control lane is: } . $arg_refs->{control_lane});

  $arg_refs->{gerald_config} = $self->_gerald_config($arg_refs);   # create config file, return its name

  my @job_ids = $self->_run_bustard($arg_refs);

  # This just makes sure the directories got created (and logs their names);

  my $gerald_dir  = $self->_find_gerald($arg_refs);           # fully qualified name
  $self->log("Gerald dir:  $gerald_dir");

  return @job_ids;

}

sub _pick_control_lane {
  my ($self) = @_;
  $self->log( q{deprecated, use 'pick_control_lane' instead} );
  return $self->pick_control_lane();
}

# ----------------------------------------------------------------------
# Check that, if the Data/Intensities folder exists (it normally
# does), it is writable.

sub _check_runfolder {

  my ($self, $arg_refs) = @_;

  my $intensity_path = $self->runfolder_path() . q{/} . $self->bustard_home();

  if ( -d $intensity_path && ! -w $intensity_path) {
    croak $intensity_path . q{ exists but is not writable};
  }

  return;

}

# ----------------------------------------------------------------------
# For version 1.6+, the index is not required in the use bases string for eland alignments, even for multiplex samples.
# This will only wok if the index part begins with an i/I

sub _corrected_for_gerald_use_bases {
  my ($self) = @_;

  my $use_bases = $self->use_bases();

  if ($self->illumina_pipeline() ne $self->casava_pipeline()) {
    $use_bases =~ s/,i\d*n*\d*//ixms;

  }

  return $use_bases;
}

# ----------------------------------------------------------------------
# Write out a Gerald config file to the runfolder directory.

sub _gerald_config {

  my ($self, $arg_refs) = @_;

  my $pipeline      = $self->illumina_pipeline();
  my $casava        = $self->casava_pipeline();
  my $run           = $self->run();
  my $run_name      = $run->name();
  my $run_folder    = $self->runfolder_path();    # fully qualified name
  my $timestamp     = $self->timestamp();
  my $config_prefix = $self->gerald_config_prefix();
  my $gerald_config = "$run_folder/$config_prefix-$timestamp.txt";
  my $use_bases     = $self->_corrected_for_gerald_use_bases;
  my $control_lane  = $arg_refs->{control_lane};

  my $email_list  = $self->emails_conf()->{gerald_email_alert};
  $email_list     =~ s{,}{ }gxms;

  my @lines;
  push @lines, "\# Gerald config file for RTA run: $run_folder\n";
  push @lines, "\# Auto-generated by: $PROGRAM_NAME v$VERSION\n";
  push @lines, "\# At: $timestamp\n";
  push @lines, "\n";
  push @lines, q{# OLB:    } . $pipeline . "\n";
  push @lines, q{# CASAVA: } . $casava   . "\n";
  push @lines, "\n";
  push @lines, "EXPT_DIR .\n";
  push @lines, "EMAIL_SERVER mail.sanger.ac.uk\n";
  push @lines, "EMAIL_DOMAIN sanger.ac.uk\n";
  push @lines, "EMAIL_LIST $email_list\n";
  push @lines, "WEB_DIR_ROOT http://sfweb.internal.sanger.ac.uk:9000/\n";
  push @lines, "\n";
  push @lines, "ELAND_MULTIPLE_INSTANCES 8\n";
  push @lines, "ELAND_SET_SIZE 8\n";          # split 2*120 tiles into batches of 8
  push @lines, "SEQUENCE_FORMAT --fastq\n";
  push @lines, "USE_BASES $use_bases\n";
#  push @lines, "ORIG_READ_LENGTHS $read_lengths\n";
  push @lines, "\n";

  foreach my $run_lane (@{$run->run_lanes()}) {

    my $library  = $run_lane->library();  # only one of $library, $control and $pool will be defined
    my $control  = $run_lane->control();
    my $pool     = $run_lane->pool();

    my $entity   = $run_lane->entity();   # entity is library || control || pool
    my $position = $run_lane->position();
    my $force_control = $self->has_control_lane_position() ? $self->control_lane_position() : 0;

    if ($control or $force_control == $position) {

      my $id = $entity->id();
      $self->log("$run_name lane $position is control id $id");
      push @lines, "\# lane $position is control id $id\n";

      my $cur_ref  = $entity->reference_sequence();

      if ( ! $cur_ref  || $self->no_eland() ) {   # if no reference, just do sequence processing

        if ( $self->no_eland() ) {
          $self->log( q{This has been set manually to run with no eland alignment} );
        } else {
          $self->log("$run_name WARNING: No reference sequence found for control $id in lane $position");
        }

        if ($self->is_paired_read()) {
          push @lines, "$position:ANALYSIS sequence_pair\n";
        } else {
          push @lines, "$position:ANALYSIS sequence\n";
        }

      } else {

        if ($self->is_paired_read()) {
          push @lines, "$position:ANALYSIS eland_pair\n";
        } else {
          push @lines, "$position:ANALYSIS eland_extended\n";
        }

        my ($control_dir) = $cur_ref =~ /(.*)\/[^\/]+$/smx;
        push @lines, "$position:ELAND_GENOME $control_dir\n";

        push @lines, "$position:QCAL_SOURCE auto\n";

      }

      push @lines, "\n";

    } else {

      if ($library) {

        my $lid = $library->id();
        $self->log("$run_name lane $position is library id $lid");
        push @lines, "\# lane $position is library id $lid\n";

      } elsif ($pool) {

        my $pid = $pool->id();
        $self->log("$run_name lane $position is pool id $pid");
        push @lines, "\# lane $position is pool id $pid\n";

      } else {
        croak "lane $position is neither library nor control nor pool";
      }

      if ($self->is_paired_read()) {
        push @lines, "$position:ANALYSIS sequence_pair\n";
      } else {
        push @lines, "$position:ANALYSIS sequence\n";
      }

      if ( $control_lane && !$self->no_eland() ) {
        push @lines, "$position:QCAL_SOURCE auto$control_lane\n";
      }

      push @lines, "\n";

    }

  }

  $self->log("$run_name Gerald config:");
  foreach my $q (@lines) {
    my $line = $q;       # $q is ref to original -- don't chomp it directly
    chomp $line;
    $self->log("     | $line");
  }

  open my $fh, '>', $gerald_config or croak "Error creating $gerald_config: $ERRNO";
  print {$fh} @lines               or croak "Error printing to $gerald_config: $ERRNO";
  close $fh                        or croak "Error closing $gerald_config: $ERRNO";

  return $gerald_config;

}

# ----------------------------------------------------------------------
# Run the Bustard+Gerald setup step to create directories and
# makefiles. It runs quickly (minutes), so no need to bsub it.

sub _run_bustard {

  my ($self, $arg_refs) = @_;
  my @job_ids;        # will be returned empty

  my $pipeline      = $self->illumina_pipeline();
  my $casava        = $self->casava_pipeline();
  my $run_folder    = $self->runfolder_path();    # fully qualified name
  my $bustard_home  = "$run_folder/" . $self->bustard_home();
  my $gerald_config = $arg_refs->{gerald_config};        # Gerald config file
  my $control_lane  = $arg_refs->{control_lane};
  my $timestamp     = $self->timestamp();
  my $bustard_out   = "$run_folder/" . $self->bustard_output_prefix() . "-$timestamp.txt";
  my $gerald_out    = "$run_folder/" . $self->gerald_output_prefix()  . "-$timestamp.txt";
  my $use_bases     = $self->_corrected_for_gerald_use_bases;
  my $read_lengths = $use_bases;
  $read_lengths =~ s/Y//gixms;
  $read_lengths =~ s/,/:/gxms;

  if (exists $ENV{'LD_LIBRARY_PATH'}) {
    $ENV{'LD_LIBRARY_PATH'} = join q{:}, $self->boost_lib(), $ENV{'LD_LIBRARY_PATH'}; ## no critic (Variables::RequireLocalizedPunctuationVars)
  } else {
    $ENV{'LD_LIBRARY_PATH'} = $self->boost_lib(); ## no critic (Variables::RequireLocalizedPunctuationVars)
  }

  my $cur_dir = getcwd();            # save for later
  if (!$ENV{TEST_BIN}) {
    chdir $run_folder or croak "could not cd to $run_folder";
  }

  my @command;
  push @command, 'LOGNAME=' . $self->runtype();
  push @command, $self->_test_bin() . "$pipeline/" . $self->bustard_executable();
  push @command, '--make --CIF';
  push @command, '--with-qval';# --with-sig2'; # turn off sig2 creation in v12.0
  push @command, '--with-second-call';
  push @command, '--no-eamss';

  # get phasing, prephasing (if reqd) and matrix that should be used for this run's analysis
  # this will pick up the user defined versions of these if applicable
  push @command, q{--phasing=} . $self->phasing_value( $control_lane );
  if ( $self->has_prephasing() ) {
    push @command, q{--prephasing=} . $self->prephasing();
  }
  push @command, q{--matrix=} . $self->matrix_value( $control_lane );

  my $tile_list = $self->tile_list();
  if ($tile_list) {
    push @command, "--tiles=$tile_list";
  }

  push @command, "$bustard_home";
  push @command, "> $bustard_out";
  push @command, '2>&1';

  my $bustard_command = join q[ ], @command;
  $self->log("Bustard command: $bustard_command");

  my $rc = system $bustard_command;
  croak "bustard command failed: $rc" if $rc;

  $self->log('bustard complete');

  # run this here, to minimise the length of time between the bustard dir being created, and being set
  # as the bustard dir to use, since if another one is modified, then this could take preference
  my $bustard_dir = $self->find_bustard($arg_refs);          # fully qualified name
  $self->make_log_dir( $bustard_dir );

  # We may have created a Data/Intensities subfolder. Try to make it
  # all-accessible.  The other possibility is that it already
  # existed. We verified that we can write to it in
  # _check_runfolder. Can't promise it's 777 though.

  my $numch = chmod $ACCESS_ALL, $bustard_home;
  if ($numch != 1) {                       # chmod returns number of files changed
    $self->log("WARNING: chmod failed for $bustard_home");
  }

  # Now do the Gerald step. As of the x.6 pipeline release, that needs
  # to be run separately from Bustard. And GERALD.pl resides in the
  # CASAVA path.

  my $length = $self->_corrected_for_gerald_use_bases();
  $length =~ s/y//gixms;

  @command = ();
  push @command, 'LOGNAME=' . $self->runtype();
  push @command, $self->_test_bin() . "$casava/" . $self->gerald_executable();
  push @command, "$gerald_config";
  push @command, "--EXPT_DIR $bustard_dir";
  push @command, '--compression gzip';
  push @command, '--make';                   # was '--FORCE'
  push @command, qq{--length $length};
  push @command, "> $gerald_out";
  push @command, '2>&1';

  my $gerald_command = join q[ ], @command;
  $self->log("Gerald command: $gerald_command");

  $rc = system $gerald_command;
  croak "gerald command failed: $rc" if $rc;

  $self->log('gerald complete');
  # try to capture the gerald directory into the run_conf file as soon as possible
  # in order to try to reduce the chance of multiple gerald directories causing problems
  my $gerald_dir = $self->_find_gerald($arg_refs);
  $self->make_log_dir( $gerald_dir );

  # IMPORTANT - this does some fixing to attempt to get gerald acting correctly for the control lane on multiplex samples
  # IT SHOULD BE REMOVED at the earliest possible convenience
  if ($self->use_bases() =~ /I/ixms) {

    $self->log(q{Correcting gerald config files for multiplex runs});

    my $config_xml = qq{$gerald_dir/config.xml};
    my $xml = slurp($config_xml);
    $xml =~ s{<ORIG_READ_LENGTHS>.*?</ORIG_READ_LENGTHS>}{<ORIG_READ_LENGTHS>$read_lengths</ORIG_READ_LENGTHS>}xms;
    open my $fh, q{>}, $config_xml or croak qq{Could not open $config_xml to write};
    print {$fh} $xml || croak qq{Unable to write corrected xml to $config_xml};
    close $fh || carp q{Could not close FileHandle};

    my $makefile_config = qq{$gerald_dir/Makefile.config};
    my $makefile = slurp($makefile_config);
    $makefile =~ s{ORIG_READ_LENGTHS:=.*?\n}{ORIG_READ_LENGTHS:=$read_lengths\n}xms;
    open my $fhmfc, q{>}, $makefile_config or croak qq{Could not open $makefile_config to write};
    print {$fhmfc} $makefile || croak qq{Unable to write corrected makefile to $makefile_config};
    close $fhmfc || carp q{Could not close FileHandle};

   $self->log(q{Corrected files});
  }

  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return @job_ids;    # it's empty, since nothing was bsub'd

}

# ----------------------------------------------------------------------
# Find the Gerald directory in the latest-created Bustard directory in
# the runfolder. Since we just created the Bustard folder, there
# should only be one Gerald folder in it.
# however, in the case that there are 2, and that one could be being modified
# as the other is created, we try to store in the run conf file

sub _find_gerald {

  my ($self, $arg_refs) = @_;

  my $run_conf = $self->run_conf();
  if ( $run_conf->{gerald_path} ) {
    $self->log( q{returning gerald_path from run conf file} );
    return $run_conf->{gerald_path};
  }


  my $bustard_dir = $self->find_bustard($arg_refs);         # fully qualified name
  my $runtype = $self->runtype();

  my @geralds = qx[ls -1td $bustard_dir/GERALD*$runtype*];
  croak "no GERALD_$runtype directories in $bustard_dir" if @geralds == 0;
  croak "more then one GERALD_$runtype directory in $bustard_dir" if @geralds > 1;
  my $gerald_dir = $geralds[0];                              # fully qualified name
  chomp $gerald_dir;

  $self->add_to_run_conf({
    gerald_path => $gerald_dir,
  });

  return $gerald_dir;

}

# ----------------------------------------------------------------------
# Create and run a bsub command for a make job from supplied parameters.

sub _bsub_command {

  my ($self, $dir, $targets, $tag, $deps) = @_;

  # $dir        directory to do make in
  # $targets    string with one or more targets to pass to make
  # $tag        starting part of jobname
  # $deps       dependencies string to be passed to bsub, including -w. may be null.

  my $directory = $dir;

  if ( ref $directory and ref$directory eq q{HASH} ) {
    $directory = $directory->{directory};
  }

  my $bsub_command = $self->_bsub_command_string($dir, $targets, $tag, $deps);

  my $cur_dir = getcwd();            # save for later
  if (!$ENV{TEST_BIN}) {
    chdir $directory or croak "could not cd to $directory";
  }

  my $job = $self->submit_bsub_command($bsub_command);

  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return $job;

}

# ----------------------------------------------------------------------
# Separated out of the submission method for testing purposes.

sub _bsub_command_string {

  my ($self, $dir, $targets, $tag, $deps) = @_;

  # $dir        directory to do make in
  # $targets    string with one or more targets to pass to make
  # $tag        starting part of jobname
  # $deps       dependencies string to be passed to bsub, including -w. may be null.

  my $position_string = q{};
  my $array_string = q{};

  if ( ref $dir and ref$dir eq q{HASH} ) {
    $deps = $dir->{depends_on};
    $targets = $dir->{target};
    $tag = $dir->{tag};
    $array_string = $dir->{array_string} || q{};
    $position_string = $dir->{position}  || q{};
    $dir = $dir->{directory};
  }

  my $run_name    = $self->run()->name();
  my $timestamp   = $self->timestamp();
  my $job_name    = ${tag} . q{_} . $timestamp;
  my $output_name = q{log/} . $job_name;
  if ( $array_string ) {
    $job_name .= $array_string;
    $output_name .= q{.%I};
  }
  $output_name .= q{.%J.out};

  my $bsub_queue  = $self->lsf_queue;
  my $jobs        = $self->processor_fork_number();  # TODO: Reduce this for longer runs

  my $nfs = $self->fs_resource;
  my $mem_req = $self->memory_requirement();

  my @command;
  push @command, 'bsub';
  push @command, "-n $jobs";
  push @command, "-q $bsub_queue";
  push @command, "-o $output_name ";
  push @command, "-J $job_name";
  push @command, '-C ' . $self->dump_size();    # get dump on failure

######  push @command, "-R \"select\[mem>$mem_req\] rusage\[mem=$mem_req\] span[hosts=1] select\[$nfs>=$jobs\] rusage\[$nfs=$jobs\]\"";
  push @command, $self->fs_resource_string( {
    resource_string => qq{-R 'select[mem>$mem_req] rusage[mem=$mem_req] span[hosts=1]'},
    counter_slots_per_job => $self->processor_fork_number(),
  });
  push @command, "-M ${mem_req}000";
  push @command, $deps || q[];

  push @command, q['];               # " enclose command in quotes
  push @command, $self->make_executable();
  push @command, "-j $jobs";
  push @command, "$targets";
  push @command, q['];               # " closing quote

  my $bsub_command = join q[ ], @command;

  $self->log("make command: $bsub_command");

  return $bsub_command;
}


=head2 make methods
# ----------------------------------------------------------------------

# The routines which follow submit 'make' jobs for post-run analysis
# as a series of steps, each dependent on the previous. At any step
# where it is possible, a separate 'make' is run for each lane, in
# parallel.

# The routines are similarly structured. Each just figures out a
# target and a tag (part of the run name), invokes _bsub_command(),
# and returns the job-id strings. Some of them have a loop over all
# the lanes.

# ----------------------------------------------------------------------

=head2 make_bustard_matrix_lanes

Submits make_bustard_matrix_lane jobs to LSF and returns the job_ids as an array

  my @job_ids = $oAnalysisBustard->make_bustard_matrix_lanes({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_bustard_matrix_lanes {

  my ($self, $arg_refs) = @_;
  my @job_ids;

  my $id_run = $self->id_run();
  my $run    = $self->run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $bustard_dir = $self->find_bustard($arg_refs);          # fully qualified name

  my @lanes;
  foreach my $lane ( @{ $run->run_lanes() } ) {
    push @lanes, $lane->position();
  }

  my $array_string = $self->create_array_string( @lanes );
  my $position = $self->lsb_jobindex();

  my $target = q{matrix_} . $position . q{_finished.txt};
  my $tag = "bustard_matrix_${id_run}";
  push @job_ids, $self->_bsub_command( {
    directory => $bustard_dir,
    target => $target,
    tag => $tag,
    depends_on => $depends_on,
    array_string => $array_string,
    position => $position,
  } );

  return @job_ids;

}

# ----------------------------------------------------------------------

=head2 make_bustard_matrix_all

Submits make_bustard_matrix_all job to LSF and returns the job_id

  my $job_id = $oAnalysisBustard->make_bustard_matrix_all({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_bustard_matrix_all {

  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $bustard_dir = $self->find_bustard($arg_refs);          # fully qualified name

  my $target = 'matrix_finished.txt';
  my $tag = "bustard_matrix_${id_run}_all";
  my $job = $self->_bsub_command( {
    directory => $bustard_dir,
    target => $target,
    tag => $tag,
    depends_on => $depends_on,
  } );

  return $job;

}

# ----------------------------------------------------------------------

=head2 make_bustard_phasing_lanes

Submits make_bustard_phasing_lane jobs to LSF and returns the job_ids as an array

  my @job_ids = $oAnalysisBustard->make_bustard_phasing_lanes({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_bustard_phasing_lanes {

  my ($self, $arg_refs) = @_;
  my @job_ids;

  my $id_run = $self->id_run();
  my $run    = $self->run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $bustard_dir = $self->find_bustard($arg_refs);          # fully qualified name

  my @lanes;
  foreach my $lane ( @{ $run->run_lanes() } ) {
    push @lanes, $lane->position();
  }

  my $array_string = $self->create_array_string( @lanes );
  my $position = $self->lsb_jobindex();

  my $target = q{phasing_} . $position . q{_finished.txt};
  my $tag = "bustard_phasing_${id_run}";
  push @job_ids, $self->_bsub_command( {
    directory => $bustard_dir,
    target => $target,
    tag => $tag,
    depends_on => $depends_on,
    array_string => $array_string,
    position => $position,
  } );

  return @job_ids;

}

# ----------------------------------------------------------------------

=head2 make_bustard_phasing_all

Submits make_bustard_phasing_all job to LSF and returns the job_id

  my $job_id = $oAnalysisBustard->make_bustard_phasing_all({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_bustard_phasing_all {

  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $bustard_dir = $self->find_bustard($arg_refs);          # fully qualified name

  my $target = 'phasing_finished.txt';
  my $tag = "bustard_phasing_${id_run}_all";
  my $job = $self->_bsub_command( {
    directory => $bustard_dir,
    target => $target,
    tag => $tag,
    depends_on => $depends_on,
  } );

  return $job;

}

# ----------------------------------------------------------------------

=head2 make_bustard_basecalls_lanes

Submits make_bustard_basecalls_lanes jobs to LSF and returns the job_ids as an array

  my @job_ids = $oAnalysisBustard->make_bustard_basecalls_lanes({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_bustard_basecalls_lanes {

  my ($self, $arg_refs) = @_;
  my @job_ids;

  my $id_run = $self->id_run();
  my $run    = $self->run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $bustard_dir = $self->find_bustard($arg_refs);          # fully qualified name

  my @lanes;
  foreach my $lane ( @{ $run->run_lanes() } ) {
    push @lanes, $lane->position();
  }

  my $array_string = $self->create_array_string( @lanes );
  my $position = $self->lsb_jobindex();

  my $target = q{s_} . $position;
  my $tag = "bustard_basecalls_${id_run}";
  push @job_ids, $self->_bsub_command( {
    directory => $bustard_dir,
    target => $target,
    tag => $tag,
    depends_on => $depends_on,
    array_string => $array_string,
    position => $position,
  } );

  return @job_ids;

}

# ----------------------------------------------------------------------

=head2 make_bustard_basecalls_all

Submits make_bustard_basecalls_all job to LSF and returns the job_id

  my $job_id = $oAnalysisBustard->make_bustard_basecalls_all({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_bustard_basecalls_all {

  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $bustard_dir = $self->find_bustard($arg_refs);          # fully qualified name

  my $target = 'all';
  my $tag = "bustard_basecalls_${id_run}_all";
  my $job = $self->_bsub_command( {
    directory => $bustard_dir,
    target => $target,
    tag => $tag,
    depends_on => $depends_on,
  } );

  return $job;

}

# ----------------------------------------------------------------------

=head2 make_gerald_tiles

Submits make_gerald_tiles job to LSF and returns the job_id

  my $job_id = $oAnalysisBustard->make_gerald_tiles({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_gerald_tiles {

  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $gerald_dir = $self->_find_gerald($arg_refs);          # fully qualified name

  my $target = 'tiles.txt';
  my $tag = "gerald_tiles_${id_run}";
  my $job = $self->_bsub_command($gerald_dir, $target, $tag, $depends_on);

  return $job;

}

# ----------------------------------------------------------------------

=head2 make_gerald_qtable

Submits make_gerald_qtable job to LSF and returns the job_id

  my $job_id = $oAnalysisBustard->make_gerald_qtable({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
    control_lane => $iControlLanePosition
  });

=cut

sub make_gerald_qtable {

  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();
  my $run    = $self->run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $gerald_dir = $self->_find_gerald($arg_refs);          # fully qualified name
  my $control_lane = $self->_pick_control_lane();

  my $target;
  if ( ! $control_lane) {
    $target = 'tiles.txt';           # this already exists => noop!
  } elsif ($run->is_paired_read()) {
    $target = "s_${control_lane}_1_qtable.txt s_${control_lane}_2_qtable.txt";
  } else {
    $target = "s_${control_lane}_qtable.txt";
  }

  my $tag = "gerald_qtables_${id_run}";
  my $job = $self->_bsub_command($gerald_dir, $target, $tag, $depends_on);

  return $job;

}

# ----------------------------------------------------------------------

=head2 make_gerald_lanes

Submits make_gerald_lanes jobs to LSF and returns the job_ids as an array

  my @job_ids = $oAnalysisBustard->make_gerald_lanes({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_gerald_lanes {

  my ($self, $arg_refs) = @_;
  my @job_ids;

  my $id_run = $self->id_run();
  my $run    = $self->run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $gerald_dir = $self->_find_gerald($arg_refs);          # fully qualified name

  my @lanes;
  foreach my $lane ( @{ $run->run_lanes() } ) {
    push @lanes, $lane->position();
  }

  my $array_string = $self->create_array_string( @lanes );
  my $position = $self->lsb_jobindex();

  my $target = q{s_} . $position;
  my $tag = "gerald_lanes_${id_run}";
  push @job_ids, $self->_bsub_command( {
    directory => $gerald_dir,
    target => $target,
    tag => $tag,
    depends_on => $depends_on,
    array_string => $array_string,
    position => $position,
  } );
  return @job_ids;
}

# ----------------------------------------------------------------------

=head2 make_gerald_all

Submits make_gerald_all job to LSF and returns the job_id

  my $job_id = $oAnalysisBustard->make_gerald_all({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub make_gerald_all {

  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();
  my $depends_on = $arg_refs->{required_job_completion};
  my $gerald_dir = $self->_find_gerald($arg_refs);          # fully qualified name

  my $target = 'all';
  my $tag = "gerald_${id_run}_all";
  my $job = $self->_bsub_command($gerald_dir, $target, $tag, $depends_on);

  return $job;

}

# ----------------------------------------------------------------------

=head2 submit_cleanup

Submit a job to invoke post-analysis processing (archival file
creation, QC, etc). All that job will do is another bsub. But this
extra level of indirection is necessary so that job dependencies can
be honoured. Returns the job_id

  my $job_id = $oAnalysisBustard->submit_cleanup({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub submit_cleanup {

  my ($self, $arg_refs) = @_;

  my $id_run = $self->id_run();
  $self->log("Submitting post-qseq job for run $id_run");

  my $bsub_command = $self->_cleanup_bsub_cmd($arg_refs);

  my $cur_dir = getcwd();            # save for later
  my $gerald_dir = $self->_find_gerald($arg_refs);          # fully qualified name
  if (!$ENV{TEST_BIN}) {
    chdir $gerald_dir or croak "could not cd to $gerald_dir";
  }

  my $job = $self->submit_bsub_command($bsub_command);

  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return $job;

}

# ----------------------------------------------------------------------
# Separated out of the submission method for testing purposes.

sub _cleanup_bsub_cmd {

  my ($self, $arg_refs) = @_;

  my $id_run      = $self->id_run();
  my $depends_on  = $arg_refs->{required_job_completion};
  my $gerald_dir  = $self->_find_gerald($arg_refs);          # fully qualified name
  my $bsub_queue  = $self->lsf_queue;
  my $timestamp   = $self->timestamp();
  my $job_name    = "post_qseq_${id_run}_${timestamp}";
  my $output_name = "log/post_qseq_${id_run}_${timestamp}-\%J.out";

  # build the analyse_rta_log_file so that this can be passed through to enable
  # the correct finding of the pipeline that was run
  $self->analyse_rta_log_file();

  $self->_set_recalibrated_path($gerald_dir);
  $self->bustard_path();
  my $options = $self->attributes_as_command_options({
    excluded_attributes => [qw(
      log_file log_file_name log_file_path subpath casava_pipeline illumina_pipeline timestamp
      bustard_home gerald_config_prefix runtype script_name time phasing prephasing matrix
      )],
  });

  my @command;
  push @command, q{bsub};
  push @command, qq{-q $bsub_queue};
  push @command, $depends_on || q[];
  push @command, $self->fs_resource_string( {
    counter_slots_per_job => 1,
  } );
  push @command, qq{-o $output_name };
  push @command, qq{-J $job_name};
  push @command, q{'} . $self->post_qseq_executable();
  push @command, $options. q{'};

  my $bsub_command = join q[ ], @command;

  $self->log("post-qseq command: $bsub_command");

  return $bsub_command;

}

# ----------------------------------------------------------------------

no Moose;
__PACKAGE__->meta->make_immutable;

1;

__END__


=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Carp

=item English -no_match_vars

=item Readonly

=item Moose

=item npg_pipeline::base

=item npg_common::roles::run::path_info

=item npg_common::roles::run::fs_resource

=item npg_pipeline::roles::business::bustard_lsf_reqs

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Tom Skelly (ts6@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
