#############
# $Id: create_psuedo_qseq_custom.pm 12510 2011-02-01 15:49:31Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-11-03
# Last Changed On: $Date: 2011-02-01 15:49:31 +0000 (Tue, 01 Feb 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-24.2/lib/npg_pipeline/analysis/create_psuedo_qseq_custom.pm $

package npg_pipeline::analysis::create_psuedo_qseq_custom;
use Moose;
use MooseX::InsideOut;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12510 $ =~ /(\d+)/mxs; $r; };
use File::Spec;

extends qw{npg_pipeline::base};
with qw{npg_common::roles::run::path_info};
with qw{npg_common::roles::run::long_info};
with qw{npg_pipeline::roles::business::flag_options};

Readonly::Scalar our $CUSTOM_SUFFIX  => q{_custom.txt};

=head1 NAME

npg_pipeline::analysis::create_psuedo_qseq_custom

=head1 VERSION

$LastChangedRevision: 12510 $

=head1 SYNOPSIS

  my $oPseudoFileMaker = npg_pipeline::analysis::create_psuedo_qseq_custom->new({
    run_folder => $sRunFolder,
    recalibrated_path => $sRecalibratedPath,
    bustard_path => $sBustardPath
  });

=head1 DESCRIPTION

This object module exposes the method run, which, should it be able to work out (or have been provided) the recalibrated path (where it should find
qseq_custom files) and the bustard_path (where the qseq files should have been created), it will check there is a qseq_custom file for each qseq file,
and any which are missing, it will create a softlink in the recalibrated path to the qseq file

npg_common::roles::run::path_info is used in this class, so you can pass through other parameters which will (within the methods exported by the role)
provide the paths to the recalibrated and bustard dirs

=head1 SUBROUTINES/METHODS

=head2 run - method to go through the qseq files in the Bustard directory, and check there is a corresponding qseq_custom file in the GERALD directory
if not, then create a softlink to the file with the qseq_custom name

  eval { $oPseudoFileMaker->run(); } or do { ... Your error handling here ... };

=cut

sub run {
  my ($self) = @_;

  my $qseq_custom_files = $self->qseq_custom_files();

  my $readname_index = $self->general_values_conf()->{qseq_readname_index};

  # 
  if ( $self->recalibrated_path() =~ /PB_cal/xms ) {
    return $self->_run_pb_cal_psuedo_custom( $qseq_custom_files );
  }

  # we will assume that if there are any qseq_custom files for a standard Illumina analysis present,
  # then they all will be, so there is no need for further deliberations
  if (scalar keys%{$qseq_custom_files}) {
    return 1;
  }


  my @qseq_files        = $self->_all_qseq_files();
  my @missing_files;
  foreach my $file (@qseq_files) {
    if (!$qseq_custom_files->{$file}) {
      push @missing_files, $file;
    }
  }

  foreach my $file (@missing_files) {
    my $orig_file = $file;

    # if run is not indexed, but file has a t qseq file, then skip, as not likely to be real
    if ( ! $self->is_indexed() && $file =~ /s_\d_t_\d+_qseq/xms ) {
      next;
    }

    # if run is indexed, and the read name is 2, then fix file to be 3
    if ( $self->is_indexed() && $file =~ /(s_\d)_2_\d+_qseq/xms ) {
      my $first_bit = $1;
      $file =~ s/${first_bit}_2/${first_bit}_3/xms;
    }

    $self->log(qq{Missing qseq_custom file for $file});
    my $file_name_to_use = $self->recalibrated_path(). qq{/$file} . q{_custom.txt};

    my $target_path = $self->control() ?  $self->qseq_location_path()
                    :                     $self->bustard_path()
                    ;

    $target_path = File::Spec->abs2rel( $target_path, $self->recalibrated_path() );
    my $cmd = q{ln -s } . $target_path . qq{/$orig_file} . qq{.txt $file_name_to_use};

    if ($self->verbose()) {
      $self->log($cmd);
    }

    my $output = qx/$cmd/;
    if ($CHILD_ERROR != 0) {
      croak qq{Error running $cmd - $EVAL_ERROR - Return Code - $CHILD_ERROR};
    }

    # readname will need fixing if it is the second end of a multiplex read
    if ($self->is_indexed() && $file =~ /s_\d_3_\d+_qseq/xms) {
      $self->log(q{Fixing readname inside file.});

      # fix the readname within the file, by creating a temporary file which will then be moved to the original
      open my $fh, q{<}, $file_name_to_use or croak qq{unable to open $file_name_to_use for reading: $EVAL_ERROR};
      my $temp_filename = q{/tmp/} . $self->id_run() . $self->lane();
      open my $temp_file, q{>}, $temp_filename or croak qq{unable to open $temp_filename to put correct readname in};
      while (my $line = <$fh>) {
        my @temp = split /\t/xms, $line;
        $temp[$readname_index] = $self->general_values_conf()->{illumina_read_two};
        print {$temp_file} join qq{\t}, @temp; ## no critic (InputOutput::RequireCheckedSyscalls)
      }
      close $fh or croak qq{unable to close $file_name_to_use from reading};
      close $temp_file or croak qq{unable to close $temp_filename from writing};

      # mv the temp_file back to the new file
      my $mv_command = qq{mv $temp_filename $file_name_to_use};
      $self->log($mv_command);
      my $return = qx[$mv_command];
      if ($CHILD_ERROR) {
        croak qq{Unable to mv back the corrected file $file_name_to_use for lane } . $self->lane();
      }

    }

    # if indexed, and we are looking at the first end file then also do the tag qseq file
    if ( $self->is_indexed() && $file =~ /(s_\d)_1_\d+_qseq/xms ) {
      my $first_bit = $1;
      $file      =~ s/${first_bit}_1/${first_bit}_2/xms;
      $orig_file =~ s/${first_bit}_1/${first_bit}_t/xms;

      my $new_file = $self->recalibrated_path(). qq{/$file} . q{_custom.txt};

      $cmd = q{ln -s } . $self->bustard_path() . qq{/$orig_file} . qq{.txt $new_file};



      if ($self->verbose()) {
        $self->log($cmd);
      }

      $output = qx/$cmd/;
      if ($CHILD_ERROR != 0) {
        croak qq{Error running $cmd - $EVAL_ERROR - Return Code - $CHILD_ERROR};
      }

      $self->log(q{Fixing readname inside file.});

      # fix the readname within the file, by creating a temporary file which will then be moved to the original
      open my $fh, q{<}, $new_file or croak qq{unable to open $new_file for reading: $EVAL_ERROR};
      my $temp_filename = q{/tmp/} . $self->id_run() . $self->lane();
      open my $temp_file, q{>}, $temp_filename or croak qq{unable to open $temp_filename to put correct readname in};
      while (my $line = <$fh>) {
        my @temp = split /\t/xms, $line;
        $temp[$readname_index] = 2;
        print {$temp_file} join qq{\t}, @temp; ## no critic (InputOutput::RequireCheckedSyscalls)
      }
      close $fh or croak qq{unable to close $new_file from reading};
      close $temp_file or croak qq{unable to close $temp_filename from writing};

      # mv the temp_file back to the new file
      my $mv_command = qq{mv $temp_filename $new_file};
      $self->log($mv_command);
      my $return = qx[$mv_command];
      if ($CHILD_ERROR) {
        croak qq{Unable to mv back the corrected file $new_file for lane } . $self->lane();
      }

    }
  }

  return 1;
}

# returns a hash of all the qseq_custom files found as keys, less the _custom.txt
has q{qseq_custom_files} => (isa => q{HashRef}, is => q{ro}, lazy_build => 1, documentation => q{Do not set on Construction - Bad things will happen});

sub _build_qseq_custom_files {
  my ($self) = @_;
  my $recalibrated_path;
  eval { $recalibrated_path = $self->recalibrated_path(); } or do { croak $EVAL_ERROR; };

  my @files = $self->_obtain_files({
    directory  => $recalibrated_path,
    file_match => q{qseq_custom.txt},
  });

  my $file_hash = {};
  foreach my $file (@files) {
    $file =~ s/_custom[.]txt.*//ixms;
    $file_hash->{$file}++;
  }

  return $file_hash;
}

# control determines if the run should, or appears to have a control lane. if it is true,
# then it will not allow copies of the qseq files in the bustard directory to be used
has q{control} => (isa => q{Bool}, is => q{ro}, lazy_build => 1);

sub _build_control {
  my ($self) = @_;

  # manual override as this run is to be processed as though there were no control lane
  if ($self->no_control_lane()) {
    return 0;
  }

  my @qtable_files = $self->_obtain_files({
    directory  => $self->recalibrated_path(),
    file_match => q{qtable.txt},
  });

  if (scalar@qtable_files) {
    return 1;
  }

  return 0;
}

# returns an array of all the qseq files, less the .txt
sub _all_qseq_files {
  my ($self, $has_control) = @_;

  my ($bustard_path, $qseq_location_path);
  eval {
    $bustard_path       = $self->bustard_path();
    $qseq_location_path = $self->qseq_location_path();
  } or do {
    croak $EVAL_ERROR;
  };

  my @files;
  eval {
    @files = $self->_obtain_files({
      directory => $qseq_location_path,
      file_match => q{qseq.txt},
    });
    scalar @files;
  } or do {
    $self->log($EVAL_ERROR);
    if ($self->control()) {
      croak q{Cannot continue, as this run has a control lane};
    }
  };

  if (!$self->control() || !scalar@files) {
    @files = $self->_obtain_files({
      directory  => $bustard_path,
      file_match => q{qseq.txt},
    });
  }

  my $file_hash;
  foreach my $file (@files) {
    $file =~ s/[.]txt.*//ixms;
    $file =~ s/_ub//ixms;
    $file_hash->{$file}++;
  }

  my @wanted_files = sort keys %{$file_hash};
  return @wanted_files;
}

# returns an array of filenames from the given directory matching the file_match string
sub _obtain_files {
  my ($self, $arg_refs) = @_;

  my $dh;
  opendir $dh, $arg_refs->{directory} or croak qq{Unable to open $arg_refs->{directory}};
  my @dir_files = readdir $dh;
  closedir $dh;

  my @wanted_files;
  foreach my $file (@dir_files) {
    if ($file =~ /\Q$arg_refs->{file_match}\E/xms) {
      push @wanted_files, $file;
    }
  }

  @wanted_files = sort @wanted_files;
  return @wanted_files;
}

# PB_cal files will be located differently, but indexed reads need copying through regardless, and anything
# which hasn't been recalibrated

sub _run_pb_cal_psuedo_custom {
  my ( $self, $qseq_custom_files ) = @_;

#foreach my $key ( sort keys %{ $qseq_custom_files } ) {
#  $self->log( $key );
#}

  my $basecalls_path = $self->basecall_path();
  my $pb_cal_path    = $self->pb_cal_path();

  my @qseq_files_in_basecalls = $self->_obtain_files( {
    directory  => $basecalls_path,
    file_match => q{_qseq.txt},
  } );

  my $missing_files = {};

  my $is_indexed = $self->is_indexed();

  $self->_truly_present_files( {
    missing_files => $missing_files,
    qseq_files_in_basecalls => \@qseq_files_in_basecalls,
    qseq_custom_files => $qseq_custom_files,
  } );

  my %replacements = (
    2 => $self->general_values_conf()->{illumina_read_two},
    t => 2,
  );

  my $number_of_missing_files = scalar keys %{ $missing_files };

  if ( ! $number_of_missing_files ) {
    $self->log( q{all files found as expected} );
    return 1;
  }

  $self->log( $number_of_missing_files . q{ missing files to psuedo down from BaseCalls} );

  foreach my $fn ( sort keys %{ $missing_files } ) {
#    next;
    my $linked_file = $basecalls_path . q{/} . $fn;

    my $link_file_base = $pb_cal_path . q{/};

    my $link_file;

    if ( ! $is_indexed ) {
      $link_file   = $link_file_base . $missing_files->{$fn} . $CUSTOM_SUFFIX;
    }

    my ( $end ) = $fn =~ /\As_\d_([t|1|2])_/xms;

    if ( ! $link_file ) {

      if ( $end eq q{1} ) {
        $link_file   = $link_file_base . $missing_files->{$fn} . $CUSTOM_SUFFIX;
      }

      if ( ! $link_file ) {

        my $temp = $missing_files->{$fn};

        # if it is an indexed(multiplexed) run, then we need to do some renaming of files
        my ($first_bit) = $temp =~ /(s_\d)_.+_\d+_qseq/xms;
        $temp =~ s/${first_bit}_2/${first_bit}_3/xms;
        $temp =~ s/${first_bit}_t/${first_bit}_2/xms;

        $link_file = $link_file_base . $temp . $CUSTOM_SUFFIX;

        # we also need to check that if read 2 has already been found (as it was recalibrated),
        # then it needs moving out of the way before the tag read takes its place
        if ( $end ne q{2} ) {
          if ( -e $link_file ) {
            my $new_file_name = $link_file;
            $new_file_name =~ s/${first_bit}_2/${first_bit}_3/xms;
            my $mv_command = qq{mv $link_file $new_file_name};
            if ( $self->verbose() ) {
              $self->log( qq{executing: $mv_command} );
            }
            my $return = qx{$mv_command};
            if ( $CHILD_ERROR ) {
              croak q{unable to rename second end : } . $return;
            }
            $self->_fix_readname_in_file( {
              file_name => $new_file_name,
              new_readname => $self->general_values_conf()->{illumina_read_two},
            } );
          }
        }
      }

    }

    my $ln_command = qq{ln -s $linked_file $link_file};
    if ( $self->verbose() ) {
      $self->log( q{link command: } . $ln_command );
    }
    my $return = qx{$ln_command};
    if ( $CHILD_ERROR ) {
      croak q{Problem creating custom qseq link: } . $return;
    }

    if ( $is_indexed && $end ne q{1} ) {
      $self->_fix_readname_in_file( {
        file_name => $link_file,
        new_readname => $replacements{$end},
      } );
    }

  }
  return 1;
}

# determine which files are present, and which are not
sub _truly_present_files {
  my ( $self, $arg_refs ) = @_;

  my $qseq_custom_files = $arg_refs->{qseq_custom_files};

  foreach my $fn ( @{ $arg_refs->{qseq_files_in_basecalls} } ) {
    my ($temp) = $fn =~ /(.*?_qseq)[.]txt/xms;
    my $comparison_temp = $temp;

    # if the read is indexed, then some funny naming has gone on as the tag read (t) was actually read 2, and read 2 (if paired) is actually read 3
    if ( $self->is_indexed() ) {
      my ( $first_bit, $end ) = $comparison_temp =~ /(s_\d)_(.)_\d+_qseq/xms;

      $comparison_temp =~ s/${first_bit}_2/${first_bit}_3/xms;
      if ( $end eq q{2} ) {
        if ( $qseq_custom_files->{$comparison_temp} ) {
#          $self->log( qq{\t$temp : $comparison_temp looks present} );
          next;
        }
        if ( $qseq_custom_files->{$temp} ) {
#          $self->log( qq{\t$temp looks present} );
          next;
        }
      }

      if ( $end eq q{t} ) {
        $comparison_temp =~ s/${first_bit}_t/${first_bit}_2/xms;
        if ( $self->is_paired_read() ) {
          my $new_temp = $temp;
          $new_temp =~ s/${first_bit}_t/${first_bit}_3/xms;
#          $self->log( $temp . q{:} . $new_temp );
          if ( $qseq_custom_files->{$new_temp} && $qseq_custom_files->{$comparison_temp} ) {
#            $self->log( qq{\tlooks all present} );
            next;
          }
          $comparison_temp = $temp; # we already now know that this is missing
        }

      }
    }

#   $self->log( q{about to decide about } . $temp . q{ : } . $comparison_temp );

    if ( ! $qseq_custom_files->{$comparison_temp} ) {
      $arg_refs->{missing_files}->{$fn} = $temp;
    }
  }
  return 1;
}

# in files, the readnames need correcting, since for GERALD and PB_cal, bit act and need demultiplexed
# readnames to match the filename end, but then SRF and fastq creation need them back to the originals
sub _fix_readname_in_file {
  my ( $self, $arg_refs ) = @_;

  my $file_name       = $arg_refs->{file_name};
  my $new_readname    = $arg_refs->{new_readname};
  my $readname_index  = $arg_refs->{readname_index} || $self->general_values_conf()->{qseq_readname_index};

  if ( ! ( $file_name || $new_readname ) ) {
    croak q{No file_name or new_readname provided};
  }

  if ( $self->verbose() ) {
    $self->log(q{Fixing readname inside file.});
  }

  open my $fh, q{<}, $file_name or croak qq{unable to open $file_name for reading: $EVAL_ERROR};
  my $temp_filename = q{/tmp/} . $self->id_run() . q{_renaming_file};
  open my $temp_file, q{>}, $temp_filename or croak qq{unable to open $temp_filename to put correct readname in};
  while (my $line = <$fh>) {
    my @temp = split /\t/xms, $line;
    $temp[$readname_index] = $new_readname;
    print {$temp_file} join qq{\t}, @temp; ## no critic (InputOutput::RequireCheckedSyscalls)
  }
  close $fh or croak qq{unable to close $file_name from reading};
  close $temp_file or croak qq{unable to close $temp_file from writing};

  # mv the temp_file back to the new file
  my $mv_command = qq{mv $temp_filename $file_name};
  if ( $self->verbose() ) {
    $self->log( $mv_command );
  }
  my $return = qx[$mv_command];
  if ($CHILD_ERROR) {
    croak qq{Unable to mv back the corrected file $file_name for lane } . $self->lane();
  }

  return 1;
}


no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
