#############
# $Id: multiplex.pm 12510 2011-02-01 15:49:31Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-09-18
# Last Changed On: $Date: 2011-02-01 15:49:31 +0000 (Tue, 01 Feb 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-24.2/lib/npg_pipeline/analysis/multiplex.pm $

package npg_pipeline::analysis::multiplex;

use Moose;
use MooseX::InsideOut;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12510 $ =~ /(\d+)/mxs; $r; };
use Perl6::Slurp;
use npg_qc::autoqc::results::tag_decode_stats;

extends q{npg_pipeline::base};
with q{npg_common::roles::run::long_info};
with q{npg_pipeline::roles::business::bustard_lsf_reqs};
with q{npg_pipeline::roles::business::flag_options};
with q{npg_pipeline::roles::business::tag_info};
with q{npg_common::roles::run::lane::tag_info};

Readonly::Scalar our $PROGRAM_FILE  => q{npg_pipeline_decode_bustard_multiplex};

=head1 NAME

npg_pipeline::analysis::multiplex

=head1 VERSION

$LastChangedRevision: 12510 $

=head1 SYNOPSIS

  my $oAnalysisMultiplex = npg_pipeline::analysis::multiplex->new({
    id_run => $iIdRun,
    run_folder => $sRunFolder,
    # ... or various paths, something here is required ...
    timestamp => $sTimeStamp, # required
  });

=head1 DESCRIPTION

Object module which knows how to deal with multiplex runs which need some additional
manipulation of files before they can be recalibrated by GERALD

=head1 SUBROUTINES/METHODS

=head2 timestamp

a string containing the timestamp for this analysis. Required on construction

  my $sTimeStamp = $oAnalysisMultiplex->timestamp();

=cut

has q{timestamp} => (isa => q{Str}, is => q{ro}, required => 1);

=head2 submit_index_decoder_jobs

Method submits to lsf index_decoder for bustard generated qseq files on indexed/multiplexed runs only.
Also creates the subfolder where the original qseqs will be stored.

  eval {
    @job_ids = $oAnalysisMultiplex->submit_index_decoder_jobs($arg_refs);
    1;
  } or do {
    ... your error handling here ...
  };

=cut

sub submit_index_decoder_jobs {
  my ($self, $arg_refs) = @_;

  # if the run is not indexed/multiplexed, there is nothing to do
  if (!$self->is_indexed()) {
    return ()
  }

  my $id_run = $self->id_run();

  $self->log(qq{Submitting Jobs to index decode qseq files for $id_run});


  $self->_create_bustard_original_qseq_dir();

  my @job_ids = ();

  $arg_refs->{array_string} = $self->create_array_string( $self->positions() );
  $arg_refs->{position} = $self->lsb_jobindex();

  push @job_ids, $self->submit_bsub_command( $self->_generate_bsub_command($arg_refs) );

  return @job_ids;
}

=head2 index_decode

Method launches index_decoder for bustard generated qseq files on indexed/multiplexed runs only and
renames the files such that GERALD doesn't need to know about indexes. The originals are stored in
another directory under Bustard, so that they aren't lost.

  eval {
    $oAnalysisMultiplex->index_decode();
  } or do {
    ... your error handling here ...
  };

=cut

sub index_decode {
  my ($self) = @_;

  $self->log(q{Index decoding for run } . $self->id_run() . q{, Lane } . $self->lane());

  $self->log(q{Moving original qseq files});

  $self->_move_original_qseqs();

  $self->log(q{Index decoding original qseq files});

  $self->_run_index_decoder();

  $self->log(q{Renaming the files to indexed names});

  $self->rename_files_to_indexed_names();

  return 1;
}


=head2 rename_files_to_indexed_names

GERALD doesn't want to know about index qseq files, so we rename them from 1,2(,3) to 1(,2),t

=cut

sub rename_files_to_indexed_names {
  my ($self) = @_;

  $self->log(q{Renaming the decoded files for lane } . $self->lane() . q{ to their indexed names});

  my $first_read_name           = $self->general_values_conf()->{sanger_read_one};
  my $index_read_original_name  = $self->general_values_conf()->{illumina_read_index};
  my $second_end_original_name  = $self->general_values_conf()->{illumina_read_two};
  my $index_read_final_name     = $self->general_values_conf()->{sanger_read_index};
  my $second_end_final_name     = $self->general_values_conf()->{sanger_read_two};

  my %new_cycle = (
      $index_read_original_name => $index_read_final_name,
      $second_end_original_name => $second_end_final_name,
    );

  my @cycles_to_rename = ($index_read_original_name);
  if ($self->is_paired_read()) {
    push @cycles_to_rename, $second_end_original_name;
  }

  my $readname_index = $self->general_values_conf()->{qseq_readname_index};

  eval {

    my $tile_count = $ENV{TEST_DIR} ? 1 : $self->tile_count();
    my @tile_names = sort keys %{ $self->lane_tile_clustercount()->{ $self->lane() } };

    if ( $ENV{TEST_DIR} ) {
      @tile_names = ( 1 );
    }

    foreach my $tile ( @tile_names ) {

      foreach my $cycle (@cycles_to_rename) {
        my $name     = $self->find_bustard() . q{/s_} . $self->lane() . q{_} . $cycle              . q{_} . (sprintf q{%04d}, $tile) . q{_qseq.txt};
        my $new_name = $self->find_bustard() . q{/s_} . $self->lane() . q{_} . $new_cycle{$cycle}  . q{_} . (sprintf q{%04d}, $tile) . q{_qseq.txt};

        if (!-e $name) {
          $self->log(qq{Cannot find $name, skipping});
          next;
        }

        # mv to rename the file
        my $mv_command = qq{mv $name $new_name};
        $self->log($mv_command);
        my $return = `$mv_command`;

        if ($CHILD_ERROR) {
          croak qq{Unable to rename tile $tile, $cycle to $tile, $new_cycle{$cycle} for lane } . $self->lane();
        }

        # fix the readname within the file, by creating a temporary file which will then be moved to the original
        open my $fh, q{<}, $new_name or croak qq{unable to open $new_name for reading: $EVAL_ERROR};
        my $temp_filename = q{/tmp/} . $self->id_run() . $self->lane();
        open my $temp_file, q{>}, $temp_filename or croak qq{unable to open $temp_filename to put correct readname in};
        while (my $line = <$fh>) {
          my @temp = split /\t/xms, $line;
          $temp[$readname_index] = $new_cycle{$cycle};
          print {$temp_file} join qq{\t}, @temp; ## no critic (InputOutput::RequireCheckedSyscalls)
        }
        close $fh or croak qq{unable to close $new_name from reading};
        close $temp_file or croak qq{unable to close $temp_file from writing};

        # mv the temp_file back to the new file
        $mv_command = qq{mv $temp_filename $new_name};
        $self->log($mv_command);
        $return = qx[mv $temp_filename $new_name];
        if ($CHILD_ERROR) {
          croak qq{Unable to mv back the corrected file $tile, $cycle to $tile, $new_cycle{$cycle} for lane } . $self->lane();
        }

      }

    }
    1;
  } or do {
    croak $EVAL_ERROR;
  };

  return 1;
}

=head2 rename_files_from_indexed_names

illumina2srf need the filenames to show the order of the run, so we rename them from 1(,2),t to 1,2(,3)
This is just a psuedo renaming, as we will actually use softlinks to do this

=cut

sub rename_files_from_indexed_names {
  my ($self) = @_;

  $self->log(q{Renaming the decoded files for lane } . $self->lane() . q{ from their indexed names});

  $self->_create_psuedo_qseq_custom_dir();

  my $first_read_name           = $self->general_values_conf()->{sanger_read_one};
  my $index_read_original_name  = $self->general_values_conf()->{illumina_read_index};
  my $second_end_original_name  = $self->general_values_conf()->{illumina_read_two};
  my $index_read_final_name     = $self->general_values_conf()->{sanger_read_index};
  my $second_end_final_name     = $self->general_values_conf()->{sanger_read_two};

  my %new_cycle = (
      $first_read_name        => $first_read_name,
      $index_read_final_name  => $index_read_original_name,
      $second_end_final_name  => $second_end_original_name,
    );

  my @cycles_to_rename = ($index_read_final_name);
  if ($self->is_paired_read()) {
    unshift @cycles_to_rename, $second_end_final_name;
  }
  unshift @cycles_to_rename, $first_read_name; # add on the first cycle so that there is a softlink for all files in the same directory

  eval {

    foreach my $position ($self->positions()) {

      foreach my $tile (1..$self->tile_count()) {

        foreach my $cycle (@cycles_to_rename) {
          my $name     = $self->recalibrated_path()       . q{/s_} . $position . q{_} . $cycle              . q{_} . (sprintf q{%04d}, $tile) . q{_qseq_custom.txt};
          my $new_name = $self->psuedo_qseq_custom_dir()  . q{/s_} . $position . q{_} . $new_cycle{$cycle}  . q{_} . (sprintf q{%04d}, $tile) . q{_qseq_custom.txt};

          my $return = qx[ln -s $name $new_name];
          if ($CHILD_ERROR && !$ENV{TEST_DIR}) {
            croak qq{Unable to softlink tile $tile, $cycle to $tile, $new_cycle{$cycle} for lane } . $self->lane() . q{ in } . $self->psuedo_qseq_custom_dir();
          }

        }

      }

    }

    1;
  } or do {
    croak $EVAL_ERROR;
  };

  return 1;
}

=head2 bustard_original_qseq_dir

location of the original qseq files created by bustard before they have been index corrected

  my $sQseqDir = $oAnalysisMultiplex->bustard_original_qseq_dir();

=cut

has q{bustard_original_qseq_dir} => (isa => q{Str}, is => q{ro}, lazy_build => 1,
  documentation => q{The directory where bustard qseq files should be put before they are index_decoded},);

sub _build_bustard_original_qseq_dir {
  my ($self) = @_;
  return $self->find_bustard() . q{/bustard_original_qseq_dir};
}

=head2 psuedo_qseq_custom_dir

location of the softlinks which allow illumina2srf to know how to handle the files by their filenames correctly

  my $sPsuedoQseqCustomDir = $oAnalysisMultiplex->psuedo_qseq_custom_dir();

=cut

has q{psuedo_qseq_custom_dir} => (isa =>=> q{Str}, is => q{ro}, lazy_build => 1,
  documentation => q{The directory where the softlinks to the qseq_custom files can be found which are named so that illumina2srf can find and use them},);

sub _build_psuedo_qseq_custom_dir {
  my ($self) = @_;
  return $self->recalibrated_path() . q{/pseudo_qseq_custom_dir};
}

# private methods

# generate the busb commands for launching the index_decoder jobs
sub _generate_bsub_command {
  my ($self, $arg_refs) = @_;

  my $required_job_completion = $arg_refs->{required_job_completion};
  my $array_string = $arg_refs->{array_string};
  my $position = $arg_refs->{position};
  my $timestamp = $self->timestamp();
  my $script_bin = $self->internal_script_bin();
  my $id_run = $self->id_run();
  my $bsub_queue  = $self->lsf_queue;
  my $resource_string = $self->fs_resource_string( {
    counter_slots_per_job => 1,
  } );

  my $bustard_path = $self->find_bustard();

  my $outfile = $self->make_log_dir( $bustard_path ) . q{/} . $self->bsub_output_file();
  my $job_name = $PROGRAM_FILE . q{_} . $id_run . q{_} . $timestamp;
  my $replacement = $job_name . q{.%I.%J};
  $outfile =~ s/substitute_this/$replacement/xms;
  $job_name .= $array_string;

  my $options = $self->attributes_as_command_options({
    excluded_attributes => [qw(
        time lsf_queue script_bin run batch default_usr_bin illumina_pipeline_bin
        summary_link max_lane owning_group mail_from control_lane_position default_control_lane
        bsub_error_file bsub_output_file subpath log_file log_file_name log_file_path script_name
        lane bustard_path
      )],
  });

  my $job_sub = qq{bsub -q $bsub_queue $required_job_completion $resource_string -J $job_name -o $outfile '$script_bin/$PROGRAM_FILE --lane $position --bustard_path $bustard_path $options'};

  if ($self->verbose()) { $self->log($job_sub); }
  return $job_sub;
}


# create the bustard_original_qseq_dir if it doesn't already exist
sub _create_bustard_original_qseq_dir {
  my ($self) = @_;

  return $self->_create_folder( $self->bustard_original_qseq_dir() );
}

# create the psuedo_qseq_custom_dir if it doesn't already exist
sub _create_psuedo_qseq_custom_dir {
  my ($self) = @_;

  return $self->_create_folder( $self->psuedo_qseq_custom_dir() );
}

# take a directory path and create that folder
sub _create_folder {
  my ($self, $folder) = @_;
  $self->log(qq{Attempting to create $folder.});

  my $return_val;
  # create the directory or croak out as we must have the directory
  if (! -d $folder) {
    $return_val = qx[mkdir $folder];
    if ($CHILD_ERROR) {
      croak qq{Unable to create folder $folder : $CHILD_ERROR};
    }
  } else {
    $self->log(q{directory already exists});
  }

  $self->log(qq{Attempting to make the $folder directory all accessible.});

  # try to make the directory all accessable. Shouldn't be an issue though, since almost all of this should be done as superuser anyway, so not fatal
  $return_val = qx[chmod 777 $folder];
  if ($CHILD_ERROR) {
    $self->log(qq{Unable to change permissions on $folder to all access.});
  }
  return 1;
}

# move all qseqs for lane to bustard_original_qseq_dir
sub _move_original_qseqs {
  my ($self) = @_;

  if (! -d $self->bustard_original_qseq_dir()) {
    $self->_create_bustard_original_qseq_dir();
  }

  my $bustard_path = $self->find_bustard();
  my $bustard_original_qseq_dir = $self->bustard_original_qseq_dir();

  my $first_read_name           = $self->general_values_conf()->{sanger_read_one};
  my $index_read_original_name  = $self->general_values_conf()->{illumina_read_index};
  my $second_end_original_name  = $self->general_values_conf()->{illumina_read_two};

  # mv the qseq files, or croak as they must be moved
  my $mv_cmd_read_one   = qq{mv $bustard_path/s_} . $self->lane() . q{_} . $first_read_name          . qq{_*_qseq.txt $bustard_original_qseq_dir/};
  my $mv_cmd_read_index = qq{mv $bustard_path/s_} . $self->lane() . q{_} . $index_read_original_name . qq{_*_qseq.txt $bustard_original_qseq_dir/};
  my $mv_cmd_read_two   = qq{mv $bustard_path/s_} . $self->lane() . q{_} . $second_end_original_name . qq{_*_qseq.txt $bustard_original_qseq_dir/};

  $self->log($mv_cmd_read_one);
  my $return_value = qx[$mv_cmd_read_one];
  if ($CHILD_ERROR) {
    croak qq{Unable to move the qseq files : $CHILD_ERROR};
  }
  $self->log($mv_cmd_read_index);
  $return_value = qx[$mv_cmd_read_index];
  if ($CHILD_ERROR) {
    croak qq{Unable to move the qseq files : $CHILD_ERROR};
  }
  if ($self->is_paired_read()) {
    $self->log($mv_cmd_read_two);
    $return_value = qx[$mv_cmd_read_two];
    if ($CHILD_ERROR) {
      croak qq{Unable to move the qseq files : $CHILD_ERROR};
    }
  }

  $self->_rename_qval_files();

  return 1;
}

# rename the qval.txt.gz files, or croak as they must be renamed
sub _rename_qval_files {
  my ($self) = @_;

  my $lane = $self->lane();

  my $bustard_path = $self->find_bustard();

  my $index_read_original_name  = $self->general_values_conf()->{illumina_read_index};
  my $second_end_original_name  = $self->general_values_conf()->{illumina_read_two};
  my $index_read_final_name     = $self->general_values_conf()->{sanger_read_index};
  my $second_end_final_name     = $self->general_values_conf()->{sanger_read_two};

  my $ls_cmd = qq{ls -1 $bustard_path/s_} . $lane . q{_} . $index_read_original_name . q{_*_qval.txt.gz*};
  my $files = qx[$ls_cmd];
  my @files = split /\n/xms, $files;

  foreach my $file (@files) {
    next if (!$file);
    $self->log($file);
    my $new_file_name = $file;
    $new_file_name =~ s/s_${lane}_${index_read_original_name}_/s_${lane}_${index_read_final_name}_/xms;
    my $mv_cmd_read_index = qq{mv $file $new_file_name};

    $self->log($mv_cmd_read_index);
    my $return_value = qx[$mv_cmd_read_index];
    if ($CHILD_ERROR) {
      croak qq{Unable to rename $file : $CHILD_ERROR};
    }
  }


  if ($self->is_paired_read()) {
    $ls_cmd = qq{ls -1 $bustard_path/s_} . $lane . q{_} . $second_end_original_name . q{_*_qval.txt.gz*};
    $files = qx[$ls_cmd];
    @files = split /\n/xms, $files;

    foreach my $file (@files) {
      next if (!$file);
      $self->log($file);
      my $new_file_name = $file;
      $new_file_name =~ s/s_${lane}_${second_end_original_name}_/s_${lane}_${second_end_final_name}_/xms;
      my $mv_cmd_read_two = qq{mv $file $new_file_name};

      $self->log($mv_cmd_read_two);
      my $return_value = qx[$mv_cmd_read_two];
      if ($CHILD_ERROR) {
        croak qq{Unable to rename $file : $CHILD_ERROR};
      }
    }

  }

  return 1;
}

sub _run_index_decoder {
  my ($self) = @_;

  my $command = $self->_index_decoder_command();

  if ( ! $command ) {
    $self->log( q{Lane } . $self->lane() . q{ has no "decode" command defined!});
    return 1;
  }

  $self->log(q{Attempting to index_decode lane } . $self->lane() . qq{ Using command:\n\n$command\n\n});

  my $return = qx[$command];
  if ($CHILD_ERROR) {
    my $error_message = qq{Error running $command: Error code: $CHILD_ERROR};
    $self->log(qq{\n$return\n$error_message});
    croak $error_message;
  }

  $self->log(qq{Output of running index_decoder:\n\n$return\n\n});

  my $tag_decode_stats_rs = npg_qc::autoqc::results::tag_decode_stats->new ({
     id_run    => $self->id_run(),
     position  => $self->lane(),
  });
  $tag_decode_stats_rs->parsing_output_string($return);
  my $output_json = $self->find_bustard() . q{/} . $self->id_run() . q{_} . $self->lane() . q{_tag_decode_stats.json};
  $tag_decode_stats_rs->store($output_json);

  return 1;
}

sub _index_decoder_command {
  my ($self) = @_;
  my $command;

  my $tag_file =  $self->get_tag_file({
      index_length => $self->index_length(),
    });

  # test to see is there is a tag, and return empty string if not
  my $tag_file_contents = slurp ( $tag_file );
  my ( $first_tag ) = $tag_file_contents =~ /([ACGT]+)/xms;

  if ( !$first_tag ) {
    $self->log( qq{$tag_file has no tags in it, returning copy instead of decoder command} );
    $command = q{cp } .
      $self->bustard_original_qseq_dir() . q{/s_} . $self->lane() . q{_?_????_qseq.txt} .
      q{ } .
      $self->find_bustard();
  }
  else{
    $self->log( qq{Found this tag first in tag_file - $first_tag. Returning a decoder command} );
    $command = $self->internal_script_bin() . q{/} . $self->external_script_names_conf()->{index_decoder};
    $command .= q{ -t } . $tag_file;

    $command .= q{ -u } . $self->corrected_tag_part_use_bases({
      tag_file => $tag_file,
    });

    $command .= q{ -o } . $self->find_bustard();
    $command .= q{ } . $self->bustard_original_qseq_dir() . q{/s_} . $self->lane() . q{_?_????_qseq.txt};
  }

  if ($self->verbose()) {
    $self->log($command);
  }
  return $command;
}

no Moose;
__PACKAGE__->meta->make_immutable;

1;

__END__


=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Carp

=item English -no_match_vars

=item Readonly

=item Moose

=item npg_pipeline::base

=item npg_common::roles::run::path_info

=item npg_common::roles::run::fs_resource

=item npg_common::roles::run::long_info

=item npg_pipeline::roles::business::bustard_lsf_reqs

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
