#############
# $Id: ClusterCounts.pm 12510 2011-02-01 15:49:31Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2010-09-07
# Last Changed On: $Date: 2011-02-01 15:49:31 +0000 (Tue, 01 Feb 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-24.2/lib/npg_pipeline/archive/file/ClusterCounts.pm $

package npg_pipeline::archive::file::ClusterCounts;
use Moose;
use MooseX::InsideOut;
use Carp;
use English qw{-no_match_vars};
use File::Spec;
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12510 $ =~ /(\d+)/mxs; $r; };
use XML::LibXML;
use List::MoreUtils qw(uniq);

extends qw{npg_pipeline::archive::file::generation};

Readonly::Scalar our $CLUSTER_COUNTS_SCRIPT => q{npg_pipeline_check_file_cluster_count};

=head1 NAME

npg_pipeline::archive::file::ClusterCounts

=head1 VERSION

$LastChangedRevision: 12510 $

=head1 SYNOPSIS


  my @job_ids;
  eval {
    my $oClusterCounts = npg_pipeline::archive::file::ClusterCounts->new({
      run_folder          => $run_folder,
      timestamp           => q{20090709-123456},
      internal_script_bin => q{/software/solexa/bin},
      id_run              => 1234,
#      verbose => 1, # use if you want logging of the commands sent to LSF
    });

    my $arg_refs = {
      required_job_completion  => q{-w'done(123) && done(321)'},
    }

    @job_ids = $oClusterCounts->launch($arg_refs);
  } or do {
    # your error handling here
  };

  my $oClusterCounts = npg_pipeline::archive::file::ClusterCounts->new({
    run_folder => $run_folder,
#    verbose => 1, # use if you want logging of the commands sent to LSF
  });
  $oClusterCounts->run_cluster_count_check();

=head1 DESCRIPTION

This module is responsible for launching a job, and then processing that job, which will check the cluster
counts are what they are expected to be in fastq, srf and other files

=head1 SUBROUTINES/METHODS

=head2 launch

Method launches lsf job commands and returns an array of job ids retrieved

  my @JobIds = $oClusterCounts->launch({
    required_job_completion  => q{-w'done(123) && done(321)'},
  });

=cut

sub launch {
  my ( $self, $arg_refs ) = @_;
  my @job_ids;

  my @positions = $self->positions( $arg_refs );

  if ( ! scalar @positions ) {
    $self->log( q{no positions found, not submitting any jobs} );
    return @job_ids;
  }

  $arg_refs->{array_string} = $self->create_array_string( @positions );

  push @job_ids, $self->submit_bsub_command( $self->_generate_bsub_command( $arg_refs ) );

  return @job_ids;
}

=head2 run_cluster_count_check

method which goes checking the cluster counts

=cut

sub run_cluster_count_check {
  my ( $self ) = @_;

  $self->_bustard_pf_cluster_count();
  $self->_bustard_raw_cluster_count();

  eval {
    $self->_check_srf();
    $self->_check_fastqcheck();
#    $self->_check_fastq(); # we don't need to do this anymore, since fastqcheck will have exactly the same number
    1;
  } or do {
    croak $EVAL_ERROR;
  };

  return 1;
}

has q{position} => (
  isa => q{Int},
  is  => q{ro},
);

#############
# private methods

has q{_bustard_pf_cluster_count} => (
  isa => q{Int},
  is  => q{ro},
  lazy_build => 1,
  writer => q{_set_bustard_pf_cluster_count},
);

has q{_bustard_raw_cluster_count} => (
  isa => q{Int},
  is  => q{ro},
  lazy_build => 1,
  writer => q{_set_bustard_raw_cluster_count},
);

sub _build__bustard_pf_cluster_count {
  my ( $self ) = @_;
  return $self->_populate_cluster_counts( q{pf} );
}

sub _build__bustard_raw_cluster_count {
  my ( $self ) = @_;
  return $self->_populate_cluster_counts( q{raw} );
}

sub _populate_cluster_counts {
  my ( $self, $type ) = @_;

  my $parser = XML::LibXML->new();

  my $xml = $parser->parse_file( $self->bustard_path() . q{/BustardSummary.xml} );

  my %h;

  my $return;

  foreach my $lx ( $xml->getElementsByTagName( q(TileResultsByLane) )->[0]->getElementsByTagName( q(Lane) ) ) {
    my $l = $lx->getElementsByTagName( q(laneNumber) )->[0]->textContent();

    if ( $l != $self->position() ) {
      next;
    }

    foreach my $rx ( $lx->getElementsByTagName( q(Read) ) ) {
      my $r = $rx->getElementsByTagName( q(readNumber) )->[0]->textContent();
      foreach my $tx ( $rx->getElementsByTagName( q(Tile) ) ) {
        foreach my $count ( qw(clusterCountRaw clusterCountPF) ) {
          $h{$l}{$count}{$r} += $tx->getElementsByTagName($count)->[0]->textContent();
        }
      }
    }
  }

  foreach my $l ( sort {$a<=>$b} keys %h ) {

    my @ccpf = uniq values %{ $h{$l}{clusterCountPF} };
    croak q(inconsistent pf counts between reads) if ( @ccpf > 1 );
    $self->_set_bustard_pf_cluster_count( $ccpf[0] );
    if ( $type eq q{pf} ) {
      $return = $ccpf[0];
    }

    my @ccr = uniq values %{ $h{$l}{clusterCountRaw} };
    croak q(inconsistent raw counts between reads) if ( @ccr > 1 );
    $self->_set_bustard_raw_cluster_count( $ccr[0] );
    if ( $type eq q{raw} ) {
      $return = $ccr[0];
    }
  }

  if ( !defined $return ) {
    croak q{Unable to determine a raw and/or pf cluster count};
  }

  return $return;

}

# generates the bsub command
sub _generate_bsub_command {
  my ($self, $arg_refs) = @_;

  my $array_string = $arg_refs->{array_string};
  my $required_job_completion = $arg_refs->{required_job_completion} || q{};
  my $timestamp = $self->timestamp();
  my $solexa_bin = $self->internal_script_bin();
  my $id_run = $self->id_run();

  my $job_name = $CLUSTER_COUNTS_SCRIPT . q{_} . $id_run . q{_} . $timestamp;

  my $archive_out = $self->archive_path() . q{/log};
  my $out_subscript = q{.%I.%J.out};
  my $outfile = File::Spec->catfile($archive_out, $job_name . $out_subscript);

  $job_name = q{'} . $job_name . $array_string . q{'};

  my $job_sub = q{bsub -q } . $self->lsf_queue() . qq{ $required_job_completion -J $job_name -o $outfile '};
  $job_sub .= File::Spec->catfile($solexa_bin, $CLUSTER_COUNTS_SCRIPT);

  $job_sub .= q{ --id_run=} . $id_run;
  $job_sub .= q{ --position=}  . $self->lsb_jobindex();
  $job_sub .= q{ --archive_path=} . $self->archive_path();
  $job_sub .= q{ --bustard_path=} . $self->bustard_path();

  if ( $self->verbose() ) {
    $job_sub .= q{ --verbose};
  }

  $job_sub .= q{'};

  if ( $self->verbose() ) {
    $self->log($job_sub);
  }

  return $job_sub;
}


# checks all fastqcheck files for this lane to ensure they have the correct count of pf reads
sub _check_fastqcheck {
  my ( $self ) = @_;

  $self->log( q{Checking fastqchecks} );
  $self->_check_fastq_or_fastqcheck(1);
  $self->log( q{fastqcheck files have the correct pf_cluster counts} );
  return 1;
}

# checks all fastq files for this lane to ensure they have the correct count of pf reads
sub _check_fastq { ## no critic (Subroutines::ProhibitUnusedPrivateSubroutines)
  my ( $self ) = @_;

  $self->log( q{Checking fastq} );
  $self->_check_fastq_or_fastqcheck();
  $self->log( q{fastq files have the correct pf_cluster counts} );
  return 1;
}

sub _check_fastq_or_fastqcheck {
  my ( $self, $check ) = @_;

  my  $expected_cluster_count = $self->_bustard_pf_cluster_count();
  $self->log(qq{Expected cluster count in fastq: $expected_cluster_count } );

  my $spiked_phix_cluster_count = $self->_check_spiked_phix_cluster_count( $check );
  if($spiked_phix_cluster_count){
     $self->log(qq{Spiked phix cluster split out: $spiked_phix_cluster_count});
     $expected_cluster_count = $expected_cluster_count - $spiked_phix_cluster_count;
     $self->log(qq{Expected cluster count in fastq after splitting spiked phix: $expected_cluster_count } );
  }

  my @filenames = $self->_fqnames( $check );

  my %split_nc_store = ();

  my $id_run = $self->id_run();
  my $position = $self->position();

  foreach my $fn ( @filenames ) {

    my $pf_count;
    if($check){
      $pf_count = $self->_fastqcheck_cluster_count($fn);
    }else{
      $pf_count = $self->_fastq_cluster_count($fn);
    }

    if ( $fn =~ /human/xms ) {
      my $first_part = $id_run . q{_} . $position;
      if ( $fn =~ /${first_part}_human/xms || $fn =~ /${first_part}_nonhuman/xms ) {
        $split_nc_store{single} += $pf_count;
      } elsif ( $fn =~ /${first_part}_1_human/xms || $fn =~ /${first_part}_1_nonhuman/xms ) {
        $split_nc_store{read_one} += $pf_count;
      } elsif ( $fn =~ /${first_part}_2_human/xms || $fn =~ /${first_part}_2_nonhuman/xms ) {
        $split_nc_store{read_two} += $pf_count;
      } else {
        $split_nc_store{read_tag} += $pf_count;
      }

    } else {
      if ( $pf_count != $expected_cluster_count ) {
        my @temp = split m{/}xms, $fn;
        my $file = pop @temp;
        my $msg = qq{$file does not contain the correct number of pf clusters\n\tExpected: $expected_cluster_count\n\tActual: $pf_count};
        $self->log( $msg );
        croak $msg;
      }
    }

  }

  foreach my $key ( keys %split_nc_store ) {
    if ( $split_nc_store{ $key } !=  $expected_cluster_count ) {
      my $msg = qq{split $key does not contain the correct number of pf clusters\n\tExpected:$expected_cluster_count \n\tActual: $split_nc_store{ $key }};
      $self->log( $msg );
      croak $msg;
    }
  }

  return 1;
}

sub _check_spiked_phix_cluster_count {
   my ($self, $check) = @_;

   my $spiked_phix_cluster_count = 0;

   my @phix_files = $self->_spiked_phix_fqnames( $check );

   if(! scalar @phix_files){
      return $spiked_phix_cluster_count;
   }

   my %counts = ();
   foreach my $phix_fn (@phix_files){
      my $count;
      if($check){
         $count = $self->_fastqcheck_cluster_count($phix_fn);
      }else {
         $count = $self->_fastq_cluster_count($phix_fn);
      }
      $counts{$count}++;
   }
   my @count_values = keys %counts;
   if(scalar @count_values > 1 ){
      my $msg = qq{Cluster count in spiked phix file are different:\n\t};
      $msg   .= join qq{\n\t}, @phix_files;
      $msg   .= qq{\n\t};
      $msg   .= q{Spiked phix fastq cluster count: }.join qq{\t}, (@count_values);
      $self->log($msg);
      croak $msg;
   }
   return $count_values[0];
}

sub _fastqcheck_cluster_count {
   my ( $self, $fastqcheck_file ) = @_;

   my $output = qx{head -n 1 $fastqcheck_file};
   my ($count) = $output =~ /(\d+)/xms;
   return $count;
}

sub _fastq_cluster_count {
   my ($self, $fastq_file) = @_;

   my $line_count = 0;
   my $pf_count = 0;

   open my $fh, q{<}, $fastq_file or croak qq{Couldn't open $fastq_file for reading'};
   while ( readline $fh ) {
      if (($line_count % $self->general_values_conf()->{lines_a_read_takes_in_fastq}) != 0) {
        $line_count++;
        next;
      }
      my $line = $_;
      chomp $line;
      if ( $line ) {
        $pf_count++;
      }
      $line_count++;
   }
   close $fh || croak qq{Unable to close filehandle for $fastq_file};

   return $pf_count;
}

# work out all the fqnames for either fastq or fastqcheck, test that the file exists in the archive dir,
# and then return any existing full paths
sub _fqnames {
  my ( $self, $check ) = @_;

  my $fastqcheck = $self->id_run() . q{_} . $self->position();

  my %filenames = (
    single_read => $fastqcheck,
    read_one => $fastqcheck . q{_1},
    read_two => $fastqcheck . q{_2},
    read_tag => $fastqcheck . q{_t},
    single_read_human => $fastqcheck . q{_human},
    read_one_human => $fastqcheck . q{_1_human},
    read_two_human => $fastqcheck . q{_2_human},
    read_tag_human => $fastqcheck . q{_t_human},
    single_read_nonhuman => $fastqcheck . q{_nonhuman},
    read_one_nonhuman => $fastqcheck . q{_1_nonhuman},
    read_two_nonhuman => $fastqcheck . q{_2_nonhuman},
    read_tag_nonhuman => $fastqcheck . q{_t_nonhuman},
  );

  my $suffix = q{.fastq};
  if ( $check ) {
    $suffix .= q{check};
  }

  my @return;

  foreach my $fn ( keys %filenames ) {
    $filenames{$fn} .= $suffix;
    my $full_path = $self->archive_path() . q{/} . $filenames{$fn};
    if ( -e $full_path ) {
      push @return, $full_path;
    }
  }

  return @return;
}

sub _spiked_phix_fqnames {
  my ( $self, $check ) = @_;

  my $fastqcheck = $self->id_run() . q{_} . $self->position();

  my %filenames = (
    single_read_phix => $fastqcheck . q{_phix},
    read_one_phix => $fastqcheck . q{_1_phix},
    read_two_phix => $fastqcheck . q{_2_phix},
    read_tag_phix => $fastqcheck . q{_t_phix},
  );

  my $suffix = q{.fastq};
  if ( $check ) {
    $suffix .= q{check};
  }

  my @return;

  foreach my $fn ( keys %filenames ) {
    $filenames{$fn} .= $suffix;
    my $full_path = $self->archive_path() . q{/} . $filenames{$fn};
    if ( -e $full_path ) {
      push @return, $full_path;
    }
  }
  return @return;
}

# check the pf read counts in srf files, incl. human/nonhuman split and spiked phix srfs if applicable
sub _check_srf {
  my ( $self ) = @_;

  $self->log( q{Checking srf cluster counts} );
  my $expected_cluster_count = $self->_bustard_pf_cluster_count();
  $self->log( qq{Expected cluster count in srf: $expected_cluster_count} );

  my $orig_srf_filename = $self->archive_path() . q{/}. $self->id_run() . q{_} . $self->position() . q{_orig.srf};
  if( -e $orig_srf_filename ){
     my $orig_srf_cluster_count = $self->_srf_cluster_count( $orig_srf_filename );
     if ( $orig_srf_cluster_count != $expected_cluster_count ) {
        my $msg = qq{$orig_srf_filename does not have the correct pf cluster count:\n\tActual: $orig_srf_cluster_count.\n\tExpected: $expected_cluster_count};
        $self->log($msg);
        croak $msg;
     }
  }

  my $phix_srf_filename = $self->archive_path() . q{/}. $self->id_run() . q{_} . $self->position() . q{_phix.srf};
  if( -e $phix_srf_filename ){
     my $phix_srf_cluster_count = $self->_srf_cluster_count( $phix_srf_filename );
     $self->log( qq{Spiked phix cluster split out:  $phix_srf_cluster_count } );
     $expected_cluster_count = $expected_cluster_count - $phix_srf_cluster_count;
     $self->log( qq{Expected cluster count in srf after splitting spiked phix: $expected_cluster_count} );
  }

  my $srf_filename = $self->archive_path() . q{/}. $self->id_run() . q{_} . $self->position() . q{.srf};
  if(! -e $srf_filename){
     $self->log( qq{no srf file $srf_filename, no checking to be done here} );
     return 1;
  }

  my $srf_cluster_count = $self->_srf_cluster_count( $srf_filename );

  if ( $srf_cluster_count != $expected_cluster_count ) {
        my $msg = qq{$srf_filename does not have the correct pf cluster count:\n\tActual: $srf_cluster_count.\n\tExpected: $expected_cluster_count};
        $self->log($msg);
        croak $msg;
  }

  ##checking nonconsented splitting correct
  $self->_srf_splitting_checking(q{human}, $expected_cluster_count);

  $self->log( q{srf files have the correct pf_cluster counts} );

  return 1;
}

sub _srf_splitting_checking {
   my ($self, $split_type, $expected_cluster_count) = @_;

   my $select_filename    = $self->archive_path() . q{/} . $self->id_run() . q{_} . $self->position() . q{_} . $split_type . q{.srf};
   my $nonselect_filename  = $self->archive_path() . q{/} . $self->id_run() . q{_} . $self->position() . q{_non} . $split_type . q{.srf};

  if ( -e $select_filename ) {

    $self->log( qq{Checking $split_type srf splitting cluster count} );
    my $select_count    = $self->_srf_cluster_count( $select_filename );
    my $nonselect_count = $self->_srf_cluster_count( $nonselect_filename );

    my $total = $nonselect_count + $select_count;

    if ( $total != $expected_cluster_count ) {
      my $msg = qq{The combined total of the $split_type and non}. $split_type. qq{ srfs do not have the correct pf cluster count\n};
      $msg   .= qq{\tExpected: } . $expected_cluster_count . qq{\n\tActual: $total};
      $self->log( $msg );
      croak $msg;
    }

    $self->log( qq{srf_counts add up to the correct total for the $split_type split files} );
  }

  return  1;
}

sub _srf_cluster_count {
   my ($self, $srf_file) = @_;

   my $srf_list_cmd = $self->_test_bin() . qq{srf_list -c $srf_file} ;
   my $cluster_count = qx{$srf_list_cmd};
   chomp $cluster_count;

   if ( $CHILD_ERROR ) {
      croak qq{Unable to run $srf_list_cmd: $CHILD_ERROR};
   }

   return $cluster_count;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=item File::Spec

=item npg_pipeline::archive

=item List::MoreUtils

=item XML::LibXML

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
