#############
# $Id: SchemaInformation.pm 12510 2011-02-01 15:49:31Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2010-09-23
# Last Changed On: $Date: 2011-02-01 15:49:31 +0000 (Tue, 01 Feb 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-24.2/lib/npg_pipeline/archive/file/SchemaInformation.pm $

package npg_pipeline::archive::file::SchemaInformation;
use Moose;
use MooseX::InsideOut;
use Carp;
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12510 $ =~ /(\d+)/mxs; $r; };
use YAML;
use File::Slurp;
use JSON;
use Cwd qw(abs_path);

use npg_common::sequence::BAM_Generation;
use npg_qc::autoqc::checks::check;
use  npg_common::Fastq_Split;

## no critic Subroutines::ProhibitUnusedPrivateSubroutines

extends qw{npg_pipeline::archive::file::generation};
with qw{
    npg_common::roles::run::intensities::config
    npg_pipeline::roles::business::bustard_lsf_reqs
  };

Readonly::Scalar our $FILE_SUFFIX => q{_schema_information.yml};
Readonly::Scalar our $SCRIPT_NAME => q{npg_pipeline_schema_information};

=head1 NAME

npg_pipeline::archive::file::SchemaInformation

=head1 VERSION

$LastChangedRevision: 12510 $

=head1 SYNOPSIS

  use npg_pipeline::archive::file::SchemaInformation;

  my $oSchemaInformation = npg_pipeline::archive::file::SchemaInformation->new({
    id_run => $iIdRun,
    archive_path => $sArchivePath,
    lane => $iLane,
    file_name => $sFileName, # only when reading, although it can work it out from the lane and id_run
  });

  $oSchemaInformation->generate_file();

=head1 DESCRIPTION

This object has methods for both trawling through the run processes used for this analysis, and outputing to
a file in the archive_path, and reading a file.

This is done on a per lane basis.

Also a method for generating a bsub command to do the trawling at the correct point in the pipeline (will
generate an array of commands for the lanes that it needs to do)

=head1 SUBROUTINES/METHODS

=head2 launch_bsub_job

=cut

sub launch_bsub_job {
  my ( $self, $arg_refs ) = @_;
  my @job_ids;

  push @job_ids, $self->submit_bsub_command( $self->_generate_bsub_command( $arg_refs ) );
  return @job_ids;
}

=head2 generate_file

This method generates a file in the given archive directory (archive_path) for the given lane

  $oSchemaInformation->generate_file();

=cut

sub generate_file {
  my ( $self ) = @_;

  $self->_build_job_details();

  $self->_write_to_file();

  return 1;
}

=head2 read_in_from_file

This method reads the data in from file from the archive directory

  my $aData = $oSchemaInformation->read_in_from_file();

=cut

sub read_in_from_file {
  my ( $self ) = @_;

  my $fn = $self->_data_filename();
  my $text = read_file( $fn );
  return Load( $text );
}

###############
# private methods

# write out the data to a file in the archive directory
sub _write_to_file {
  my ( $self, $information ) = @_;

  $information ||= [ $self->_all_program_details() ];

  if ( ! ref $information ) {
    croak qq{$information is not a reference};
  }

  my $fn = $self->_data_filename();
  write_file( $fn, Dump $information );

  return 1;
}

#####################
# controlling methods
# build up the program details by looping through the data in the log files
sub _build_job_details {
  my ( $self ) = @_;

  my %script_runner = (
    analyse_rta => q{npg_pipeline_analyse_RTA},
    pb_cal      => q{npg_pipeline_PB_cal},
    post_qseq   => q{npg_pipeline_post_qseq},
  );

  $self->_rta();

  # for any analysis, we will have 2 of the 3 in this list
  foreach my $process ( qw{ analyse_rta pb_cal post_qseq } ) {

    my $log_file_method = $process . q{_log_file};
    my $has_method = q{has_} . $log_file_method;

    if ( $self->$has_method() ) {

      # get in the version of the pipeline part used
      $self->_add_program_details( {
        ID => $process,
        PN => $script_runner{ $process },
        VN => $self->_pipeline_script_job_version( $script_runner{ $process } ),
      } );

      # get in all other programs
      $self->_loop_through_jobs( $self->$log_file_method() );
    }

  }

  return 1;
}

# obtain the version from the pipeline script version run
sub _pipeline_script_job_version {
  my ( $self, $prog_name ) = @_;

  my $cmd = $self->filesystem_locations_conf()->{bin} . q{/} . $prog_name . q{ --version};
  my $version_string = qx{$cmd};
  my ( $version ) = $version_string =~ /version\s+(\d+)\n/xms;
  $version =~ s/\s//gmxs;
  return $version;
}

# go through a log files array of jobs and for each, add the version... information
sub _loop_through_jobs {
  my ( $self, $run_log ) = @_;

  my $log_hash = $self->_parse_run_log( $run_log );

  my $previous_program;
  foreach my $function_hash ( @{ $log_hash->{functions} } ) {
    my $method = q{_} . $function_hash->{function};

    # either there is a method which will determine information about a program run
    # or it is probably an internal part, and just needs recording
    if ( $self->can( $method ) ) {
      $self->$method( $previous_program );
    } else {
      $self->_add_program_details( {
        PP => $previous_program,
        PN => $function_hash->{function},
        ID => $function_hash->{function},
        VN => undef,
      } );
    }

    $previous_program = $function_hash->{function};
  }

  return 1;
}

# parse the run log and get return the json section at the end as a data structure,
# which will describe all launched jobs
sub _parse_run_log {
  my ( $self, $run_log ) = @_;

  my $file_content = read_file( $run_log );

  # Grab the json string
  #

  my ( $json_string ) = $file_content =~ /\A.*JSON[ ]Dispatch[ ]Tree:\s+(.*?)\s+\[\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\]/xms;
  if ( !$json_string ) {
    $file_content =~ s/\A.*JSON[ ]Dispatch[ ]Tree:\s+//xms;
    $file_content =~ s/\s+\[\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\][ ]Unsetting[ ][$]ENV[{]NPG_WEBSERVICE_CACHE_DIR[}]//xms;
    $json_string = $file_content;
  }

  $json_string =~ s/\s*\z//xms;

  return from_json( $json_string );
}

# information store
has q{_information} => (
  traits        => ['Array'],
  isa           => q{ArrayRef[HashRef]},
  is            => q{ro},
  default       => sub { [] },
  handles       => {
    _add_program_details   => q{push},
    _first_program_details => q{shift},
    _all_program_details   => q{elements},
    _count_program_details => q{count},
    _clear_program_details => q{clear},
  },
);

# the filename for storing/reading the data
sub _data_filename {
  my ( $self ) = @_;
  return $self->archive_path() . q{/} . $self->id_run() . q{_} . $self->lane() . $FILE_SUFFIX;
}

##############
# primary pipeline

# get RTA software version number from config xml file
# also, if this is a pb_cal pipeline, then add in the basecalling as well
sub _rta {
  my ( $self ) = @_;

  my $rta = {
    ID => q{SCS},
    VN => $self->rta_version(),
    PN => $self->rta_name(),
    DS => q{Controlling software on instrument},
  };

  $self->_add_program_details( $rta );

  if ( $self->has_pb_cal_log_file() ) {
    my $basecall = {
      ID => q{basecalling},
      VN => $self->rta_version(),
      PN => $self->rta_name(),
      DS => q{Basecalling package used on instrument (RTA)},
    };

    $self->_add_program_details( $basecall );

  }

  return $rta;
}

# Illumina primary pipeline

# get bustard software version number from config xml file
sub _run_bustard {
  my ( $self ) = @_;

  my $bustard = {
    ID => q{basecalling},
    DS => q{the basecalling package(OLB - Bustard) package used},
  };

  my $bustard_path = $self->bustard_path();
  my $bustard_config_xml = $bustard_path . q{/config.xml};

  if ( ! -e $bustard_config_xml ) {
    carp qq{Bustard config xml NOT exist: $bustard_config_xml};
    return $bustard;
  }

  my $dom = XML::LibXML->new->parse_file( $bustard_config_xml );
  my $ele;
  my @childnodes = $dom->getElementsByTagName('Run')->[0]->childNodes();
  foreach my $child (@childnodes){
     if($child->nodeName() eq q{Software}){
        $ele = $child;
        last;
     }
  }
  
  my $version = $ele->getAttribute('Version');
  my $name = $ele->getAttribute('Name');

  $bustard->{PN} = $name;
  $bustard->{VN} = $version;

  $self->_add_program_details( $bustard );

  return $bustard;
}

# get gerald software version number from config xml file
sub _make_gerald_tiles {
  my ( $self ) = @_;

  # gerald would not be run if there was no recalibration done on the lane
  if ( $self->no_recalibration() || $self->no_control_lane() ) {
    $self->log( q{No recalibration performed on lane} . $self->lane() . q{, so no Gerald data recorded} );
    return {};
  }

  if ( ! $self->pick_control_lane() ) {
    $self->log( q{No control lane found, so lane} . $self->lane() . q{ cannot have been recalibrated, so no Gerald data recorded} );
  }

  my $gerald = {
    ID => q{recalibration},
    PN => q{Gerald},
    DS => q{the recalibration CASAVA(Gerald) package used},
  };
  my $gerald_path = $self->recalibrated_path();
  my $gerald_config_xml = $gerald_path . q{/config.xml};

  if ( ! -e $gerald_config_xml ) {
    carp qq{Gerald config xml NOT exist: $gerald_config_xml};
    return $gerald;
  }

  my $dom = XML::LibXML->new->parse_file( $gerald_config_xml );
  my $bustard_version = $dom->getElementsByTagName('SOFTWARE_VERSION')->[0]->textContent();
  my $casava_version  = $dom->getElementsByTagName('BIN_DIR')->[0]->textContent();

  $gerald->{VN} = $bustard_version . qq{($casava_version)};

  $self->_add_program_details( $gerald );

  return $gerald;
}

# PB_cal primary pipeline

# bcl
sub _run_bcl_setup {
  my ( $self ) = @_;

  my $bcl_to_qseq = $self->external_script_names_conf()->{bcl_to_qseq};

  my $cmd = q{head -n 1 } . $self->basecall_path() . q{/log/qseq_files_} . $self->id_run() . q{_all_*.out};
  my $output = qx{$cmd};
  if ( $CHILD_ERROR ) {
    croak q{Unable to obtain bcl version: } . $output;
  }

  my ( $version ) = $output =~ m{/(OLB.*?)/}xms;

  my $href = {
    ID => q{bcl_to_qseq},
    PN => $bcl_to_qseq,
    VN => $version . q{ (OLB)},
    DS => q{package used to convert binary call files to qseq text files},
  };

  $self->_add_program_details( $href );
  return $href;
}

# pb_cal (harold) alignment
sub _harold_alignment_files {
  my ( $self, $previous_program ) = @_;

  $self->_control_recalibration_done(0); # ensure that this boolean is set to 0 before we start going through this

  # harold alignment would not be run if there was no recalibration done on the lane
  if ( $self->no_recalibration() ) {
    $self->log( q{No recalibration performed on lane} . $self->lane() . q{, so no harold_alignment data recorded} );
    return {};
  }

  my $alignment_script = $self->pb_cal_pipeline_conf()->{alignment_script};

  my $log_file = $self->pb_cal_path() . q{/log/PB_cal_align_} . $self->id_run() . q{_} . $self->lane() . q{_*.out};
  my $file_list = qx{ls -t $log_file};
  my @files;
  if($file_list){
    @files = split /\n/xms, $file_list;
    $log_file = shift @files;
  }

  my $output_head = q{};
  my $output_tail = q{};

  if ( -e $log_file  ) {

    my $cmd = q{head -n 15 } . $log_file;
    $output_head = qx{$cmd};
    if ( $CHILD_ERROR ) {
      croak q{Unable to read alignment log: } . $output_head;
    }

    $cmd = q{head -n 100 } . $log_file;
    $output_tail = qx{$cmd};
    if ( $CHILD_ERROR ) {
      croak q{Unable to read alignment log: } . $output_tail;
    }

  } else {
    $self->_control_recalibration_done(1);
  }

  my @temp = split /\n/xms, $output_head;
  my %info_pairs;
  foreach my $row ( @temp ) {
    my ( $key, $value ) = split /\s+/xms, $row;
    $info_pairs{$key} = $value;
  }

  my $description = q{Performs an alignment of the lane's reference genome against the data, ready to produce a calibration table; Ref: };
  if ( ! $info_pairs{reference} ) {
    $self->log( q{No reference for alignments for lane} . $self->lane() );
    $description .= q{no reference (need to use a control if found);}
  } else {
    $description .= q{Lanes own reference;};
  }

  my $href = {
    ID => q{recalibration_alignment},
    PN => $alignment_script,
    DS => $description,
    CL => ($self->_parse_for_command_line( $output_tail ) || q{not run}),
  };

  $self->_add_program_details( $href );
  return $href;
}

# pb_cal (harold) calibration table
sub _harold_calibration_tables {
  my ( $self, $previous_program ) = @_;

  # harold calibration would not be run if there was no recalibration done on the lane
  if ( $self->no_recalibration() ) {
    $self->log( q{No recalibration performed on lane} . $self->lane() . q{, so no harold_calibration_tables data recorded} );
    return {};
  }

  my $cmd = q{ls -t } . $self->recalibrated_path() . q{/log/PB_cal_table_} . $self->id_run() . q{_} . $self->lane() . q{_*.out};
  my $file_list = qx{$cmd};
  my @files = split /\n/xms, $file_list;

  my $output = q{};
  my $command_line = q{};

  my $making_cal_table;

  my $additional_criterion = q{};

  my $first_file = 1;

  foreach my $file ( @files ) {

    if ( $first_file ) {
      $output = qx{head -n 15 $file};
      if ( $CHILD_ERROR ) {
        croak q{unable to read } . $file . q{: } . $output;
      }

      my @temp = split /\n/xms, $output;
      foreach my $row ( @temp ) {
        if ( $row =~ /filter[ ]tiles/ixms ) {
          $additional_criterion .= $row . q{;};
        }
        if ( $row eq q{Making calibration table} ) {
          $making_cal_table++;
        }
      }
      $first_file = 0;
    }

    $output = qx{tail -n 100 $file};
    if ( $CHILD_ERROR ) {
      croak q{unable to read } . $file . q{: } . $output;
    }

    $command_line .= $self->_parse_for_command_line( $output );

  }

  my $description = q{Creates the calibration table for the lane based on the alignment done; Lane calibration table made: };

  if ( ! $making_cal_table ) {
    $self->log( q{No lane calibration table created for } . $self->lane() );
    $self->_control_recalibration_done(1);
    $description .= q{False;};
  } else {
    $description .= q{True;};
    $description .= $additional_criterion;
  }

  my $cal_table_script = $self->pb_cal_pipeline_conf()->{cal_table_script};

  my $href = {
    ID => q{recalibration_table_creation},
    PN => $cal_table_script,
    DS => $description,
    CL => ( $command_line || q{not run} ),
  };

  $self->_add_program_details( $href );
  return $href;
}

has q{_control_recalibration_done} => (
  isa => q{Int},
  is  => q{rw},
  default => 0,
);

# pb_cal (harold) recalibration
sub _harold_recalibration {
  my ( $self, $previous_program ) = @_;

  # harold recalibration would not be run if there was no recalibration done on the lane
  if ( $self->no_recalibration() ) {
    $self->log( q{No recalibration performed on lane} . $self->lane() . q{, so no harold_recalibration data recorded} );
    return {};
  }

  my $cmd = q{ls -t } . $self->recalibrated_path() . q{/log/PB_cal_score_} . $self->id_run() . q{_} . $self->lane() . q{_*.out};
  my $file_list = qx{$cmd};
  my @files = split /\n/xms, $file_list;

  my $output = q{};
  my $used_cal_table;
  my %info_pairs;

  my $command_line = q{};

  my $first_file = 1;

  foreach my $file ( @files ) {

    if ( $first_file ) {

      $output = qx{head -n 15 $file};
      if ( $CHILD_ERROR ) {
        croak q{unable to read } . $file . q{: } . $output;
      }

      my @temp = split /\n/xms, $output;
      foreach my $row ( @temp ) {

        next if !$row;

        my ( $key, $value ) = split /\s+/xms, $row;
        $info_pairs{$key} = $value;
        if ( $row =~ /\Areading[ ]calibration[ ]table/xms ) {
          $used_cal_table = $row;
          $used_cal_table =~ s/\Areading[ ]calibration[ ]table\s+//xms;
          $used_cal_table =~ s/\s+\z//xms;
        }
      }
      $first_file = 0;
    }

    $output = qx{tail -n 100 $file};
    if ( $CHILD_ERROR ) {
      croak q{unable to read } . $file . q{: } . $output;
    }

    $command_line .= $self->_parse_for_command_line( $output );
  }

  

  my $description = q{Recalibrates the quality scores based on the calibration table used; Calibration table used: };

  if ( ! $used_cal_table ) {
    $self->log( q{No recalibration performed on lane} . $self->lane() . q{, so no harold_recalibration data recorded, as no cal table used} );
    $description .= q{No Recalibration performed;};
  } elsif ( $used_cal_table eq $info_pairs{ct} && ! $self->_control_recalibration_done() ) {
    $description .= q{Lane;};
  } else {
    $description .= q{Control;};
  }

  my $recalibration_script = $self->pb_cal_pipeline_conf()->{recalibration_script};

  my $href = {
    ID => q{recalibration},
    PN => $recalibration_script,
    DS => $description,
    CL => ( $command_line || q{not run} ),
  };

  $self->_add_program_details( $href );
  return $href;
}

# index_decoder (or equivalent) header hashref
# this only adds anything if the lane itself is a pool (multiplexed)
# put below both primary pipelines, as used by both before the recalibration steps
sub _demultiplex {
  my ( $self ) = @_;

  if ( ! $self->is_indexed() ) {
    $self->log( q{Run} . $self->id_run() . q{ is not indexed, so no demultiplexing data recorded} );
    return {};
  }

  if ( ! $self->is_multiplexed_lane( $self->lane() ) ) {
    $self->log( q{Lane} . $self->lane() . q{ is not indexed, so no demultiplexing data recorded} );
    return {};
  }

  my $demultiplexing_script = $self->external_script_names_conf()->{index_decoder};
  my $cmd = $self->filesystem_locations_conf()->{bin} . q{/} . $demultiplexing_script . q{ -v};

  my $help_output = qx{$cmd};

  my ( $version ) = $help_output =~ /${demultiplexing_script}\s+(.*?)\n/xms;

  my $demultiplex = {
    ID => q{demultiplex},
    PN => $demultiplexing_script,
    VN => $version,
    DS => q{the script which processes tag sequences to determine the sample that the sequence belongs to},
  };

  $self->_add_program_details( $demultiplex );

  return $demultiplex;
}

###############
# secondary pipeline

# qseq2fastq (or equivalent) header hashref
sub _create_fastq {
  my ( $self, $previous_program ) = @_;

  my $cmd = $self->filesystem_locations_conf()->{bin} . q{/} . $self->external_script_names_conf()->{fast_q_creator_script} . q{ -help};

  my $help_output = qx{$cmd};

  my $archive_path = $self->archive_path();

  $cmd = qq{ls -t $archive_path/log/fastq_creation_} . $self->id_run() . q{_*.} . $self->lane() . q{.*.out};
  my $file_list = qx{$cmd};
  my @files = split /\n/xms, $file_list;

  my $output = q{};
  my $command_line = q{};

  if ( @files ) {
    my $recalibrated_path = $self->recalibrated_path();
    $output = qx{tail -n 100 $files[0]};
    $output =~ s/${recalibrated_path}\///gxms;
    $output =~ s/--unique\s+s_.*qseq_custom[.]txt/--unique <qseq_files>/gxms;
    $command_line = $self->_parse_for_command_line( $output );
    $command_line =~ s{[ ]>[ ].*?fastq}{ > <fastq_name>}xms;
    $command_line =~ s/--md5[ ].*?fastq[.]md5/--md5 <md5_name>/xms;
    $command_line =~ s/--fastqcheck[ ].*?fastqcheck/--fastqcheck <fastqcheck_name>/xms;
  }

  my ( $version ) = $help_output =~ /version\s+(.*?)\n/xms;
  $version =~ s/\s//gxms;

  my $fastq_creation = {
    ID => q{fastq_creator},
    PN => $self->external_script_names_conf()->{fast_q_creator_script},
    VN => $version,
    PP => $previous_program,
    DS => q{the program which converts the qseq files to fastq files},
    CL => ( $command_line || q{not_run} ),
  };

  $self->_add_program_details( $fastq_creation );

  return $fastq_creation;
}

# fastq_split non_consented module version number
sub _split_nonconsented_sequence {
  my ( $self ) = @_;

  if ( $self->_does_lane_need_splitting( $self->lane() )  ne q{true}) {
    $self->log( q{Lane } . $self->lane() . q{ does not need to have human split out, so no split_nonconsented_sequence data recorded} );
    return {};
  }

  my $archive_path = $self->archive_path();
  my $ls_cmd = qq{ls -t $archive_path/log/split_nonconsented_} . $self->id_run() . q{_*.} . $self->lane() . q{.*.out};
  my $file_list = qx{$ls_cmd};
  my @files = split /\n/xms, $file_list;

  my $output = q{};
  my $command_line = q{};

  if ( @files ) {
    $output = qx{tail -n 100 $files[0]};
    $command_line = $self->_parse_for_command_line( $output );
  }

  my $version = npg_common::Fastq_Split->VERSION;

  my $href = {
    ID => q{split_nonconsented_sequence},
    PN => q{fastq_split},
    VN => $version,
    CL => $command_line,
    DS => q{splits out non_consented DNA sequences from the sequenced data},
  };

  $self->_add_program_details( $href );

  return $href;
}

sub _split_phix {
  my ( $self ) = @_;

  if (  $self->_is_lane_spiked_phix( $self->lane() ) ne 'true' ) {
    $self->log( q{Lane } . $self->lane() . q{ does not need to have spiked phix split out, so no split_phix data recorded} );
    return {};
  }

  my $archive_path = $self->archive_path();
  my $ls_cmd = qq{ls -t $archive_path/log/split_phix_} . $self->id_run() . q{_*.} . $self->lane() . q{.*.out};
  my $file_list = qx{$ls_cmd};
  my @files = split /\n/xms, $file_list;

  my $output = q{};
  my $command_line = q{};

  if ( @files ) {
    $output = qx{tail -n 100 $files[0]};
    $command_line = $self->_parse_for_command_line( $output );
    my ($first_part) = $command_line =~ /(^.+?)&&\s+mv/mxs;
    $first_part =~ s/\s+$//;
    if($first_part){
      $command_line = $first_part;
    }
  }

  my $version = npg_common::Fastq_Split->VERSION;

  my $href = {
    ID => q{split_spiked_phix},
    PN => q{fastq_split},
    VN => $version,
    CL => $command_line,
    DS => q{splits out spiked phix out from the sequenced data},
  };

  $self->_add_program_details( $href );

  return $href;
}

# split_fastq_by_tag module version number
sub _split_fastq_by_tag {
  my ( $self ) = @_;

  if ( ! $self->is_indexed() ) {
    $self->log( q{Run} . $self->id_run() . q{ is not indexed, so no split_fastq_by_tag data recorded} );
    return {};
  }

  if ( ! $self->is_multiplexed_lane( $self->lane() ) ) {
    $self->log( q{Lane} . $self->lane() . q{ is not indexed, so no split_fastq_by_tag data recorded} );
    return {};
  }

  my $version = npg_common::sequence::SplitFastqByTag->VERSION;

  my $href = {
    ID => q{split_fastq_by_tag},
    PN => q{split_fastq_by_tag},
    VN => $version
  };

  $self->_add_program_details( $href );

  return $href;
}

# fastqcheck (or equivalent) header hashref
sub _create_fastqcheck {
  my ( $self, $previous_program ) = @_;

# at the moment, there is no version number reported by fastqcheck
#  my $cmd = $self->filesystem_locations_conf()->{bin} . q{/} . $self->external_script_names_conf()->{fast_qcheck_creator_script} . q{ -help};

#  my $help_output = qx{$cmd};

#  my ( $version ) = $help_output =~ /version\s+(.*?)\n/xms;
#  $version =~ s/\s//gxms;

  my $fastqcheck_creation = {
    ID => q{fastqcheck_creator},
    PN => $self->external_script_names_conf()->{fast_qcheck_creator_script},
    PP => $previous_program,
#    VN => $version,
  };

  $self->_add_program_details( $fastqcheck_creation );

  return $fastqcheck_creation;
}

# illumina2srf (or equivalent) header hashref
sub _create_srf {
  my ( $self, $previous_program ) = @_;

  if ( $self->no_srf_generation() ) {
    $self->log( q{no srf generation run for lane} . $self->lane() . q{, so no srf_creation data recorded} );
    return {};
  }

  my $srf_creator_script = $self->external_script_names_conf()->{srf_creator_script};
  my $cmd = $self->filesystem_locations_conf()->{bin} . q{/} . $srf_creator_script . q{ -help};

  my $help_output = qx{$cmd};

  my ( $version ) = $help_output =~ /${srf_creator_script}\s+v(.*?)\n/xms;
  $version =~ s/\s//gxms;

  my $srf_creation = {
    ID => q{srf_creator},
    PN => $srf_creator_script,
    VN => $version,
    PP => $previous_program,
  };

  $self->_add_program_details( $srf_creation );

  return $srf_creation;
}

# md5sum or equivalent
sub _create_md5 {
  my ( $self, $previous_program ) = @_;

  my $md5sum = $self->external_script_names_conf()->{md5_creator_script};
  my $cmd = $self->filesystem_locations_conf()->{user_bin} . q{/} . $md5sum . q{ --version};

  my $output = qx{$cmd};

  my ( $version ) = $output =~ /${md5sum}\s+[(]GNU[ ]coreutils[)]\s+(.*?)\n/xms;
  $version =~ s/\s//gxms;

  my $md5_creation = {
    ID => q{md5_creator},
    PN => $md5sum,
    VN => $version,
    PP => $previous_program,
  };

  $self->_add_program_details( $md5_creation );

  return $md5_creation;
}

# qc

has q{autoqc_checks_check} => (
  is  => q{ro},
  isa => q{Object},
  lazy_build => 1,
);

sub _build_autoqc_checks_check {
  my ( $self ) = @_;
  return npg_qc::autoqc::checks::check->new( {
    position => $self->lane(),
    id_run   => $self->id_run(),
    path     => $self->runfolder_path(),
  } );
}

sub _qc_qX_yield { ## no critic (NamingConventions::Capitalization)
  my ( $self, $previous_program ) = @_;
  return $self->_qc_run( $previous_program, q{qX_yield} );
}

sub _qc_insert_size {
  my ( $self, $previous_program ) = @_;
  return $self->_qc_run( $previous_program, q{insert_size} );
}

sub _qc_adapter {
  my ( $self, $previous_program ) = @_;
  return $self->_qc_run( $previous_program, q{adapter} );
}

sub _qc_contamination {
  my ( $self, $previous_program ) = @_;
  return $self->_qc_run( $previous_program, q{contamination} );
}

sub _qc_sequence_error {
  my ( $self, $previous_program ) = @_;
  return $self->_qc_run( $previous_program, q{sequence_error} );
}

sub _qc_gc_fraction {
  my ( $self, $previous_program ) = @_;
  return $self->_qc_run( $previous_program, q{gc_fraction} );
}

sub _qc_gc_bias {
  my ( $self, $previous_program ) = @_;
  return $self->_qc_run( $previous_program, q{gc_bias} );
}

sub _qc_run {
  my ( $self, $previous_program, $id ) = @_;

  my $qc_script = $self->external_script_names_conf()->{qc_script};

  my $cmd = $self->filesystem_locations_conf()->{bin} . q{/} . $qc_script . q{ --version};

  my $output = qx{$cmd};

  my ( $version ) = $output =~ /version\s+(.*?)\n/xms;
  $version =~ s/\s//gxms;

  my $qc_wrapper = {
    ID => $id,
    PN => $qc_script,
    VN => $version,
    PP => $previous_program,
  };

  $self->_add_program_details( $qc_wrapper );

  return $qc_wrapper;
}

####################
# bam fields - this isn't important for the bam header, since it is better placed to get more accurate data
# but is here so that there is an accurate as possible record in the schema file

# sam header object - this may not be needed
has q{_bam_generation_object} => (
  isa => q{npg_common::sequence::BAM_Generation},
  is  => q{ro},
  lazy_build => 1,
);

sub _build__bam_generation_object {
  my ( $self ) = @_;

  return $self->new_with_cloned_attributes( q{npg_common::sequence::BAM_Generation}, {
    fastq_files => [],
    bam_filename => q{},
    position => $self->lane(),
  } );
}

# bam generation
sub _bam_generation {
  my ( $self ) = @_;

  if ( $self->no_bam_generation() ) {
    $self->log( q{no bam generation run for lane} . $self->lane() . q{, so no bam creation wrapper data recorded} );
    return {};
  }

  my $fq2bam_wrapper = $self->external_script_names_conf()->{bam_creator_script};

  my $cmd = $self->filesystem_locations_conf()->{bin} . q{/} . $fq2bam_wrapper . q{ --version};

  my $help_output = qx{$cmd};

  my ( $version ) = $help_output =~ /version\s+(.*?)\n/xms;
  $version =~ s/\s//gxms;

  my $bam_creation_wrapper = {
    ID => q{fastq_to_bam_wrapper},
    PN => $fq2bam_wrapper,
    VN => $version,
    DS => q{the program used to convert fastq to bam},
  };

  $self->_add_program_details( $bam_creation_wrapper );

  # record the information about the various parts that would be used for bam generation
  $self->_bwa();
  $self->_fastq2sam();
  if ( -e $self->bustard_path() . q{/SecondCall} ) {
    $self->_sam_second_call();
  }
  $self->_sam_header();
  $self->_samtools();
  $self->_picard();
  $self->_sam_index_tag();

  return $bam_creation_wrapper;
}

# bam_markduplicate script
sub _bam_markduplicate {
  my ( $self ) = @_;

  if ( $self->no_bam_generation() ) {
    $self->log( q{no bam generation run for lane} . $self->lane() . q{, so no bam mark duplicate wrapper data recorded} );
    return {};
  }

  my $bam_markduplicate_script = $self->external_script_names_conf()->{bam_markduplicate_script};

  my $cmd = $self->filesystem_locations_conf()->{bin} . q{/} . $bam_markduplicate_script . q{ --version};

  my $help_output = qx{$cmd};

  my ( $version ) = $help_output =~ /version\s+(.*?)\n/xms;
  $version =~ s/\s//gxms;

  my $bam_markduplicate_wrapper = {
    ID => q{bam_markduplicate_wrapper},
    PN => $bam_markduplicate_script,
    VN => $version,
    DS => q{the script used to determine duplicate reads within the bam file},
  };

  $self->_add_program_details( $bam_markduplicate_wrapper );

  return $bam_markduplicate_wrapper;
}

# bwa header hashref
sub _bwa {
  my ( $self ) = @_;

  my $bwa_version = $self->autoqc_checks_check->aligner_version( $self->_bam_generation_object()->aligner_cmd() );

  my $bwa = {
    ID => q{bwa},
    PN => q{bwa},
    VN => $bwa_version,
    DS => q{bwa used for alignments and bam generation},
  };

  # get the command line options used
  eval {
    $bwa->{CL} = $self->_bam_generation_object()->aligner_options();
    1;
  } or do {
    $self->log( $EVAL_ERROR );
    $bwa->{CL} = q{unable to obtain command line options at time of running};
  };

  $self->_add_program_details( $bwa );
  return $bwa;
}

# fastq2sam version
sub _fastq2sam {
  my ( $self ) = @_;
  my $version = npg_common::sequence::Fastq2SAM->VERSION;
  my $href = {
    ID => q{fastq2sam},
    PN => q{fastq2sam},
    VN => $version,
    DS => q{the program which converts fastq to sam},
  };
  $self->_add_program_details( $href );
  return $href;
}

# sam_second_call version
sub _sam_second_call {
  my ( $self ) = @_;
  my $version = npg_common::sequence::SAM_Second_Call->VERSION;
  my $href = {
    ID => q{sam_second_call},
    PN => q{sam_second_call},
    VN => $version,
    DS => q{code used to obtain and use the second base call},
  };
  $self->_add_program_details( $href );
  return $href;
}

# sam_header version
sub _sam_header {
  my ( $self ) = @_;
  my $version = npg_common::sequence::SAM_Header->VERSION;
  my $href = {
    ID => q{sam_header},
    PN => q{sam_header},
    VN => $version,
    DS => q{ code used to generate the header in a sam file},
  };
  $self->_add_program_details( $href );
  return $href;
}

# samtool version
sub _samtools {
  my $self = shift;

  my $samtools_version = $self->autoqc_checks_check()->aligner_version( $self->_bam_generation_object()->samtools_cmd() );
  my $href = {
    ID => q{samtools},
    PN => q{samtools},
    VN => $samtools_version,
    DS => q{samtools used},
  };
  $self->_add_program_details( $href );
  return $href;
}

# picard version
sub _picard {
  my ( $self ) = @_;

  my ($picard_path) = grep {/picard/mxs} split /\s/mxs, $self->_bam_generation_object()->picard_converter_cmd();
  $picard_path = abs_path $picard_path;
  my ($picard_version) = $picard_path =~ /picard-tools-(\d+\.\d+)/mxs;

  my $href = {
    ID => q{picard},
    PN => q{picard},
    VN => $picard_version,
    DS => q{picard used},
  };
  $self->_add_program_details( $href );
  return $href;
}

# sam_index_tag script version
sub _sam_index_tag {
  my ( $self ) = @_;

  my $version =npg_common::sequence::SAM_Index_Tag->VERSION;
  my $href = {
    ID => q{sam_index_tag},
    PN => q{sam_index_tag},
    VN => $version,
    DS => q{the sam index tag code used},
  };

  $self->_add_program_details( $href );
  return $href;
}

# common code to fish out the command line from a log file
# (note, if many jobs write to the same log file, then it will only get the first one)
sub _parse_for_command_line {
  my ( $self, $string ) = @_;

  if(!$string){
    return;
  }

  my $lane = $self->lane();

  my ($command_line) = $string =~ /User[ ]input(.*?)-----------------/xms;
  $command_line =~ s/\n//gxms;
  $command_line =~ s/\t//gxms;
  my $lsb_jobindex = $self->lsb_jobindex();
  my $position_decode_string = $self->position_decode_string();
#  my $tag_index_decode_string = $self->tag_index_decode_string();
#croak $lsb_jobindex . q{:} . $lane;
  $command_line =~ s/\Q$lsb_jobindex\E/$lane/gxms;
  $command_line =~ s/\Q$position_decode_string\E/$lane/gxms;
  return $command_line;
  
}

#############
# bsub creation
sub _generate_bsub_command {
  my ( $self, $arg_refs ) = @_;

  my $required_job_completion = $arg_refs->{required_job_completion};
  my $timestamp   = $self->timestamp();
  my $script_bin  = $self->internal_script_bin();
  my $id_run      = $self->id_run();

  my $outfile     = $self->make_log_dir( $self->archive_path() ) . q{/} . $self->bsub_output_file();
  my $job_name    = q{create_schema_of_run_} . $id_run . q{_} . $timestamp;
  my $replacement = $job_name . q{.%I.%J};
  $outfile        =~ s/substitute_this/$replacement/xms;
  $job_name .= $self->create_array_string( $self->positions( $arg_refs ) );

  # build the post_qseq log file so it can be passed through for reading
  $self->post_qseq_log_file();

  my $options = $self->attributes_as_command_options({
    excluded_attributes => [ qw(
      log_file log_file_name log_file_path subpath casava_pipeline illumina_pipeline timestamp
      bustard_home gerald_config_prefix runtype script_name time bsub_output_file lane
      ) ],
  });

  my $job_sub = q{bsub -q } . $self->lsf_queue() . q{ };

  $job_sub .=  ( $self->fs_resource_string( {
    counter_slots_per_job => 1,
  } ) ) . q{ };

  $job_sub .= qq{$required_job_completion -J $job_name -o $outfile '$script_bin/} . $SCRIPT_NAME . q{ } . $options;

  $job_sub .= q{ --lane } . $self->lsb_jobindex() . q{'};

  if ( $self->verbose() ) { $self->log( $job_sub ); }
  return $job_sub;

}

###############
# These are not private methods as such, but should be considered almost as such, since they are only there to
# allow an aligner_cmd and samtools_cmd to be passed through if they really are expected to be different from the norm

has q{aligner_cmd} => (
  isa => q{Str},
  is => q{ro},
  documentation => q{If the aligner command that is used for bam generation is different from the usual, then you can feed this in here},
);

has q{samtools_cmd} => (
  isa => q{Str},
  is => q{ro},
  documentation => q{If the samtools command that is used for bam generation is different from the usual, then you can feed this in here},
);



no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=item YAML

=item File::Slurp

=item npg_common::sequence::BAM_Generation

=item npg_common::roles::run::intensities::config

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
