#############
# $Id: Spidering.pm 12510 2011-02-01 15:49:31Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2010-06-15
# Last Changed On: $Date: 2011-02-01 15:49:31 +0000 (Tue, 01 Feb 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-24.2/lib/npg_pipeline/roles/business/Spidering.pm $

package npg_pipeline::roles::business::Spidering;
use Moose::Role;
use Carp;
use English qw{-no_match_vars};
use List::MoreUtils qw(none);
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12510 $ =~ /(\d+)/mxs; $r; };
use npg_common::request;

with qw{npg_common::roles::test_type_of_value};

=head1 NAME

npg_pipeline::roles::business::Spidering

=head1 VERSION

$LastChangedRevision: 12510 $

=head1 SYNOPSIS

  package My:Spidering::Packaage;
  use Moose;
  with qw{npg_pipeline::roles::business::Spidering};

=head1 DESCRIPTION

Provides options and intelligence for spidering out of information.
The objective of which is to get the data discovered in a serialised
way and then presumably, if the option is there, get it cached, so
that when the pipeline wants it again, it is retrieved from any cache
files.

=head1 SUBROUTINES/METHODS

=head2 save2cache

boolean attribute which tells constructs an npg::api::util object for objects that tell it to save the data it retrieves to
a cache on the filesystem, which is dependent on $ENV{NPG_WEBSERVICE_CACHE_DIR} having an real directory path in it

=cut

has q{save2cache} => (
  isa => q{Bool},
  is => q{ro},
  predicate => q{has_save2cache},
);

=head2 retrieve_all_npg

Spiders out and retrieves all npg related information given a npg::api class name and it's primary key
It returns the initial top level object, or croaks at any given time.

=cut

sub retrieve_all_npg {
  my ( $self, $primary_key, $class ) = @_;

  eval { require $self->convert_package_name_to_filename( $class ); } or do { croak $EVAL_ERROR; };

  my $primary_key_name;
  my @fields = $class->fields();
  if ( $fields[0] ) {
    $primary_key_name = shift @fields;
  } else {
    my @class_split = split /::/xms, $class;
    my $type = pop @class_split;
    $primary_key_name = q{id_} . $type;
  }

  my $attrs = {
    $primary_key_name => $primary_key,
  };
  # we may need to add a special useragent, in which case, this will do so
  $self->_add_repository_and_useragent( {
    href => $attrs,
    util => 1,
  } );

  my $object;

  eval {

    # npg run_status_dicts are almost always liekly to be needed somewhere
    require npg::api::run_status_dict;
    my $rsd = npg::api::run_status_dict->new( $attrs );
    $rsd->run_status_dicts();

    $object = $class->new( $attrs );

    # first port of call to spidering is over fields
    if ( $object->can( q{fields} ) ) {
      $self->_get_fields( $object, q{npg} );
    }

    if ( $object->can( q{run_lanes} ) ) {
      my @lpositions = $self->positions;
      foreach my $run_lane ( @{ $object->run_lanes() } ) {
        my $lane_pos = $run_lane->position;
        if (none { $lane_pos == $_ } @lpositions) {
	  next;
	}

        $self->_get_fields( $run_lane, q{npg} ); # get all fields populated, most likely will already be so

        $self->_get_fields( $run_lane->entity(), q{npg} ); # most lane stuff is entity based, so go and get some of this

        $run_lane->contains_unconsented_human();

        $self->_get_references( $run_lane ); # generally, something is likely to want a reference sequence out of this, so go and get it
      }
    }

    if ( $object->can( q{instrument} ) ) {
      $self->_get_fields( $object->instrument(), q{npg} );
    }

    1;
  } or do {
    croak q{Problem obtaining NPG information: } . $EVAL_ERROR;
  };

  return $object;
}

=head2 retrieve_all_sequencescape

Spiders out and retrieves all st related information given a st::api class name and it's primary key
It returns the initial top level object, or croaks at any given time.

=cut

sub retrieve_all_sequencescape {
  my ( $self, $primary_key, $class ) = @_;

  eval { require $self->convert_package_name_to_filename( $class ); } or do { croak $EVAL_ERROR; };

  # generally, the primary key field is the first one of the fields list, but if there is no list,
  # then create most likely name, which is the format 
  my $primary_key_name;
  my @fields = $class->fields();
  if ( $fields[0] ) {
    $primary_key_name = shift @fields;
  } else {
    my @class_split = split /::/xms, $class;
    my $type = pop @class_split;
    $primary_key_name = $type . q{_id};
  }

  my $attrs = {
    $primary_key_name => $primary_key,
  };

  if ( $self->has_useragent() ) {
    $attrs->{useragent} = $self->useragent();
  }

  $self->_add_repository_and_useragent( {
    href => $attrs,
    util => 1,
  } );

  my @lpositions = $self->positions;
  my $object = $class->new( $attrs );

  # first port of call to spidering is over fields
  eval {
    # first port of call to spidering is over fields
    if ( $object->can( q{fields} ) ) {
      $self->_get_fields( $object, q{sequencescape} );
    }

    if ( $object->can( q{lanes} ) ) {
      foreach my $lane ( @{ $object->lanes() } ) {

        my $lane_pos = $lane->position;
        if (none { $lane_pos == $_ } @lpositions) {
	        next;
	      }

        if ( $lane->can( q{fields} ) ) {
          $self->_get_fields( $lane, q{sequencescape} );
        }

        my $entity = $lane->entity();
        my $h = $entity->required_fragment_size();
        foreach my $asset_id ( keys %{ $h } ) {
          my $sample;
          my $study;
          eval {
            require st::api::asset;
            my $asset = st::api::asset->new( {
              util      => $lane->util(),
              useragent => $lane->useragent(),
              id        => $asset_id,
            } );
            $asset->name();
            $sample = $asset->sample();
            if ( ! $sample ) {
              $self->log( qq{No sample obtained from asset $asset_id - not croaking});
            } else {
              $sample->name();
              $study = $sample->study();
              if ( ! $study ) {
                $self->log( q{No study obtained from sample } . $sample->id() . q{ - not croaking});
              } else {
                $study->name();
              }
            }
	        } or do {
	          $self->log( qq{Error Retrieving assets ($asset_id), related to required fragment size: } . $EVAL_ERROR );
	          croak $EVAL_ERROR;
	        };
        }
      }

    }
    1;
  } or do {
    croak q{Problem obtaining Sequencescape information: } . $EVAL_ERROR;
  };

  return $object;
}

=head2 convert_package_name_to_filename

converts a package name (My::Package:Name) to the corresponding filename (My/Package/Name.pm)
for use in require

  my $filename = $self->convert_package_name_to_filename( $package_name );

=cut

sub convert_package_name_to_filename {
  my ( $self, $package ) = @_;

  $package =~ s{::}{/}gxms;
  if ( $package !~ /[.]pm\z/xms ) {
    $package .= q{.pm};
  }

  return $package;
}

=head2 ref_repository

attribute which can set and override the base root for finding the repository where references are kept
don't set this unless you really mean it

=cut

has q{ref_repository} => (
  isa => q{Str},
  is => q{ro},
  predicate => q{has_ref_repository},
  documentation => q{Only set this if you want the reference repository root to be different from the standard},
);

=head2 useragent
=head2 npg_api_util

If you need to do some special things, then you can set pass in these objects. An npg_api_util object will be created if save2cache
has been set to true, or if you have passed in a useragent

Be warned, you would do well to leave these alone if you don't know why you should set them (and possibly, even if you do :) )

=cut

# useragent and util - should only be used if needed
has q{useragent} => (
  isa => q{Object},
  is  => q{rw},
  predicate => q{has_useragent},
);

has q{npg_api_util} => (
 isa => q{Object},
 is => q{rw},
 lazy_build => 1,
);

###########
# private methods

# some objects will call back to objects already retrieved once. We don't want to infinitely call the same stuff over and over
has q{_seen_that_already} => (
  isa => q{HashRef},
  is => q{rw},
  default => sub { return {}; },
);

sub _already_seen {
  my ( $self, $object ) = @_;

  my $seen = $self->_seen_that_already();

  # for most objects with a fields array, the first item is a primary key accessor
  if ( $object->can( q{fields} ) ) {

    my @fields = $object->fields();
    my $primary_key = shift @fields;

    my $pk_value = $object->$primary_key();
    my $class = ref $object;
    if ( ! $pk_value ) {
      return 0;
    }
    if ( $seen->{$class}->{$pk_value} ) {
      return 1;
    }

    $seen->{$class}->{$pk_value}++;
  }

  return 0;
}

# ensures that we retrieve all expected info
sub _get_fields {
  my ( $self, $object, $source ) = @_;

  if ( $self->_already_seen( $object ) ) {
    return;
  }

  if (ref $object eq q[st::api::lane]) {
    my @positions = $self->positions;
    my $position = $object->position;
    if (none { $position == $_ } @positions) {
      return;
    }
  }

  foreach my $field ( $object->fields() ) {

    my $value;
    eval {
      $value = $object->$field();
      1; # $value might legitimately be a false value, including undef
    } or do {
      # it looks as though some older stuff isn't pushed through any more as it was never used in the first place
      if ( ! ( $EVAL_ERROR && $EVAL_ERROR =~ /Can[']t[ ]call[ ]method[ ]["]getFirstChild["]/xms ) ) {
        croak $EVAL_ERROR;
      }
    };

    if ( $self->is_object( $value ) ) {
      if ( $value->can( q{fields} ) ) {
        $self->_get_fields( $value, $source ); # if it is an object, and that object has fields, then loop through them
      }
    }
  }
  if ( $source eq q{sequencescape} ) {
    my $extra_methods = $self->_sequencescape_methods_to_test();

    my $pkg = ref $object;
    my @temp = split /::/xms, $pkg;
    my $type = pop @temp;

    my $additional_methods_to_test = $extra_methods->{$type};

    foreach my $method ( @{ $self->_sequencescape_methods_to_test()->{$type} || [] } ) {

      if ( $object->can( $method ) ) {

        my $return = $object->$method();

        if ( $self->is_object( $return ) ) {
          $self->_get_fields( $return, $source );
        } elsif ( $self->is_array( $return ) ) {
          if ( $return->[0] && $self->is_object( $return->[0] ) ) {
            foreach my $array_object ( @{ $return } ) { ## no critic (ControlStructures::ProhibitDeepNests)
              $self->_get_fields( $array_object, $source );
            }
          }
        }
      }
    }
  }
  return;
}

sub _build_npg_api_util {
  my ( $self ) = @_;

  my $attrs = {};

  if ( $self->has_useragent() ) {
    $attrs->{useragent} = $self->useragent();
  }

  if ( $self->has_save2cache() ) {
    $attrs->{save2cache} = $self->save2cache();
  }

  require npg::api::util;
  return npg::api::util->new( $attrs );
}

sub _sequencescape_methods_to_test {
  my ( $self ) = @_;
  return {

    asset => [ qw{
      sample
      children
      parents
      requests
      studies
      required_fragment_size
    } ],
    request => [ qw{
      asset
      target_asset
      libraries
      project
    } ],
    lane => [ qw{
      entity
    } ],
    batch => [ qw{
      lanes
    } ],
   };
}


# ensure any calls made include looking for stuff to help obtain reference paths

sub _get_references {
  my ( $self, $run_lane ) = @_;

  require npg_common::sequence::reference;
  my $attrs = {
    id_run => $run_lane->id_run(),
    position => $run_lane->position(),
  };

  # this is probably likely to be used mainly for testing, although you could do a bit of overriding here
  $self->_add_repository_and_useragent( {
    href => $attrs,
    api_util => 1,
    ref_repository => 1,
  } );

  my $ref;
  eval {
    $ref = npg_common::sequence::reference->new( $attrs );
    $ref->refs();
    1;
  } or do {}; # don't worry to much here at the moment just make sure that messages from reference finder come back to us

  foreach my $msg ( $ref->messages()->messages() ) {
    if ( $self->can( q{log} ) ) {
      $self->log( $msg );
    } else {
      carp $msg;
    }
  }

  return 1;
}

# if you want to add your own repository and/or useragent to a href
# this is quite likely if you have set save2cache, as it is probably the only way to get it through
sub _add_repository_and_useragent {
  my ( $self, $arg_refs ) = @_;

  if ( $arg_refs->{util} || $arg_refs->{api_util} ) {
    if ( $self->save2cache() || $self->has_useragent() || $ENV{npg_common::request::cache_dir_var_name()} ) {
      if ( $arg_refs->{api_util} ) {
        $arg_refs->{href}->{api_util} = $self->npg_api_util();
      }
      if ( $arg_refs->{util} ) {
        $arg_refs->{href}->{util} = $self->npg_api_util();
      }
    }
  }

  if ($arg_refs->{ref_repository} ) {
    if ( $self->has_ref_repository() ) {
      $arg_refs->{href}->{ref_repository} = $self->ref_repository();
    }
  }

  return 1;
}

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose::Role

=item Carp

=item English -no_match_vars

=item Readonly

=item List::MoreUtils

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
